/****************************************************************/
/*								*/
/* const.h:	some constants and ext's for all		*/
/*								*/
/*		Design: Walter Benzing 1994			*/
/*		NO RIGHTS RESERVED				*/
/*								*/
/****************************************************************/

#ifndef _CONST_H
#define _CONST_H

/****************************************************************/
/*								*/
/* The next lines describe Data types that NEED a specific	*/
/* number of BITS, they are HARDWARE dependent and might be	*/
/* changed according to YOUR SYSTEM				*/
/*								*/
/****************************************************************/

#define myINT32 int			/* 32 Bit integer */
#define myINT16	short			/* 16 Bit integer */
#define myINT8	char			/* 8 bit integer */

#define PI 3.14159265358979323846
#define LN10 2.302585092994
#define MIN_FEATURE 1e-5
#define OK 0
#define WARN 1
#define ERROR 2
#define TRUE 1
#define FALSE 0
#define MAX_VIEWPORT_NUMBER 32		/* maximum number of viewports */
#define MAX_COL_NUMBER 256		/* maximum number of colors used,
					   if > 256, in Imakefile -DD16 or -DD32
					   must be used,also it must be smaller than
					   MAX_DECO_NUMBER in coor.h*/
#define MAX_MAPPED 1024			/* maximum number of mapping entries */
#define MAP_INDEX 262144		/* calculate map index */
#define MAX_NAME_LENGTH 256		/* number of charcters per name */
#define MAX_LINK_LENGTH 32		/* number of char's per link*/
#define NUM_PREF 15			/* number of preferences */
#define NUM_DATA 9			/* number of entries in data file */
#define CHAR2BITS 3			/* shift the sizeof values to get depth in bits */
#define CHARBITS 8			/* number of bits per character */
#define MAX_DIM_LENGTH 32		/* maximum character representation in char */
#define NOCHANGE 0
#define CHANGED 1
#define TRANS_CHANGED 2			/* color translation changed */
#define TRANS_AND_CHANGE (TRANS_CHANGED|CHANGED) 
#define DATA_CHANGED 4			/* data changed */
#define DATA_AND_CHANGE (DATA_CHANGED|CHANGED)
#define ALL_CHANGED (CHANGED | TRANS_CHANGED | DATA_CHANGED)
#define FILE_SEP '/'
#define ALLOC_POINTS 50
#define CUT_START 1
#define CUT_END 2
#define VALID_CUT 3
#define CUT_WINDOW 4
#define ALIGN_LEFT 1
#define ALIGN_CENTER_X 2
#define ALIGN_RIGHT 4
#define ALIGN_TOP 8
#define ALIGN_CENTER_Y 16
#define ALIGN_BOTTOM 32
#define POS_TOP_X 1
#define POS_TOP_Y 2
#define COLOR_MATCH 4
#define COLOR_3D 3
#define COLOR_3DLINE 1
#define DECO_REDRAW 100			/* Time between two decoding redraws */
#define MAX_POLY 1024			/* Maximum Number of points per Polyg */
#define XD_COMMENT 0
#define XD_POLY	1			/* XDim 3D Data types */
#define XD_LINE 2
#define XD_ELLIPS 3
#define XD_ROT 4
#define XD_3D_DATA 5
#define XD_RGB 6
#define XD_ZCOL 7
#define XD_ONE_IND 8
#define XD_PER_VALUE 9
#define XD_3D_ONE_IND	1
#define XD_3D_RGB	2			/* 3D options for import & export */
#define XD_3D_ZCOL	4			/* color index = zvalue */
#define XD_3D_PER_VALUE	8
#define XD_SORTXY	0
#define XD_SORT		16
#define XD_SORTXZ	32
#define XD_SORTYZ	64
#define MAX_INSIDE	10

/************* Data Types **************************************/

#define MY_INT -1
#define MY_DOUBLE 0
#define MY_UCHAR 1
#define MY_TRIANGLE 2

/************* Redraw Methods by Name **************************/

#define RED_GRID	"Grid"
#define RED_LINES	"Lines"
#define RED_HIDDEN	"Hidden"
#define RED_FILLED	"Filled"
#define RED_HFILLED	"HFilled"
#define RED_CUT		"Cut"
#define RED_COMP	"Complex"
#define RED_HCOMP	"HComplex"
#define RED_3DGRID	"3DGrid"
#define RED_3DLINES	"3DLines"
#define RED_3DFILLED	"3DFilled"
#define RED_3DHFILLED	"3DHFilled"

/***************************************************************/

#ifdef _MAIN

const int Short2Char = ((sizeof(short)-1) << CHAR2BITS);
const char TypeNames[2][8] = { "double", "uchar" };
const char RMethods[12][12] = {RED_GRID, RED_LINES, RED_HIDDEN, RED_FILLED,
	RED_HFILLED, RED_CUT, RED_COMP, RED_HCOMP, RED_3DGRID, RED_3DLINES,
	RED_3DFILLED, RED_3DHFILLED};
int CUChar = MY_UCHAR;
int CDouble = MY_DOUBLE;

#else

extern const int Short2Char;
extern const char TypeNames[2][8];
extern const char RMethods[12][12];
extern int CUChar;
extern int CDouble;

#endif /* _MAIN */

#endif /* _CONST_H */
