/****************************************************************/
/*								*/
/* file.c:	file I/O					*/
/*								*/
/*		Design: Walter Benzing 1994			*/
/*								*/
/*		NO RIGHTS RESERVED				*/
/****************************************************************/

/* standard includes */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <setjmp.h>
#include <math.h>

/* my includes */

#include "xdim.h"

char *FI_StripFilename(char *fileName)
   {
    int i, len;
    char *cp;
    
    len = strlen(fileName);
    cp = fileName + len;
    for(i=0; i< len; i++)
       {
	if(*(--cp) == FILE_SEP)
	  break;
       }
    return(cp);
   }

/*
 * Save a preference file
 */

int FI_SavePref(XDimInfo *XDim)
   {
    FILE *out;
    XDColorInfo *XDC;
    int i;

    XDC = &(XDim->colors);
    out = fopen(XDC->pref, "w");
    if(out)
      {
       fprintf(out, "XDim preference file:\n");
       fprintf(out, "Window width: %d\n", XDC->wwidth);
       fprintf(out, "Window height: %d\n", XDC->wheight);
       fprintf(out, "Picture width: %d\n", XDC->pwidth);
       fprintf(out, "Picture height: %d\n", XDC->pheight);
       fprintf(out, "Fast Redraw: %s\n",
	RMethods[XT_GetRedrawIndex((funcCast)(XDC->FastRedraw), XDim)]);
       fprintf(out, "Normal Redraw: %s\n",
	RMethods[XT_GetRedrawIndex((funcCast)(XDC->NormalRedraw), XDim)]);
       fprintf(out, "Interpolate Colors: %d\n", XDC->interCol);
       fprintf(out, "Default foreground: %d\n", XDC->fore);
       fprintf(out, "Default background: %d\n", XDC->back);
       fprintf(out, "Text color: %d\n", XDC->textFore);
       fprintf(out, "Axis color: %d\n", XDC->axisFore);
       fprintf(out, "Axis background: X: %d   Y: %d   Z: %d\n",
		XDC->axisBackX, XDC->axisBackY, XDC->axisBackZ);
       fprintf(out, "Time interval: %ld\n", XDC->tInterval);
       fprintf(out, "Deco area from %d to %d\n", XDC->colS, XDC->colE);
       fprintf(out, "Color rows: %d\n", XDC->rowNum);
       for(i=0; i<XDC->rowNum; i++)
          {
           fprintf(out,"%d. START:  R: %d   G: %d   B: %d\n",
             (i+1),(XDC->beginR)[i], (XDC->beginG)[i], (XDC->beginB)[i]);
           fprintf(out,"   END:  R: %d   G: %d   B: %d   Depth: %d\n",
             (XDC->endR)[i], (XDC->endG)[i], (XDC->endB)[i], (XDC->colors)[i]);
           
          }
       fclose(out);
       return(OK);
      }
    else
      {
       sprintf(XDim->error, XDERR_OPEN_WRITE_S, XDC->pref);
       return(ERROR);
      }
   }

/*
 * Load a preference file
 */

int FI_LoadPref(XDimInfo *XDim)
   {
    FILE *in;
    XDColorInfo *XDC;
    XDHardwareInfo *HWI;
    char buf[MAX_NAME_LENGTH];
    char FRedraw[32], NRedraw[32];
    int i,dummy, rowNum, row, wanted, depth,status;
    int colS, colE, fore, back, textFore, axisFore, axisBX, axisBY, axisBZ;
    unsigned short sR, sG, sB, eR, eG, eB;
    unsigned long int tInterval;
    int interCol;
    void *redfunc;

    XDC = &(XDim->colors);
    HWI = &(XDim->hwInfo);
    in = fopen(XDC->pref, "r");
    if(in)
      {
       fgets(buf, MAX_NAME_LENGTH, in);
       if(strncmp(buf, "XDim preference file:\n", MAX_NAME_LENGTH))
	 {
	  strcpy(XDim->error, XDERR_NO_PREF);
	  fclose(in);
          return(ERROR);
         }
       i=0;
       status = OK;
       fore = XDC->fore;
       back = XDC->back;
       textFore = XDC->textFore;
       axisFore = XDC->axisFore;
       axisBX = XDC->axisBackX;
       axisBY = XDC->axisBackY;
       axisBZ = XDC->axisBackZ;
       tInterval = XDC->tInterval;
       colS = XDC->colS;
       colE = XDC->colE;
       rowNum = 0;
       interCol = XDC->interCol;
       strcpy(FRedraw,RMethods[XT_GetRedrawIndex((funcCast)(XDC->FastRedraw),
	XDim)]);
       strcpy(NRedraw,RMethods[XT_GetRedrawIndex((funcCast)(XDC->NormalRedraw)
       , XDim)]);
       while(i<NUM_PREF && fgets(buf, MAX_NAME_LENGTH, in))
          {
	   if(*buf != '#')
	     {
	      if(!sscanf(buf, "Window width: %d", &(XDC->wwidth))
	        &&!sscanf(buf, "Window height: %d", &(XDC->wheight))
	        &&!sscanf(buf, "Picture width: %d", &(XDC->pwidth))
	        &&!sscanf(buf, "Picture height: %d", &(XDC->pheight))
		&&!sscanf(buf, "Fast Redraw: %s", FRedraw)
		&&!sscanf(buf, "Normal Redraw: %s", NRedraw)
		&&!sscanf(buf, "Interpolate Colors: %d", &interCol)
	        &&!sscanf(buf, "Default foreground: %d", &fore)
	        &&!sscanf(buf, "Default background: %d", &back)
		&&!sscanf(buf, "Text color: %d", &textFore)
		&&!sscanf(buf, "Axis color: %d", &axisFore)
		&&!sscanf(buf, "Axis background: X: %d   Y: %d   Z: %d",
		&axisBX, &axisBY, &axisBZ)
		&&!sscanf(buf, "Time interval: %ld", &tInterval)
		&&sscanf(buf, "Deco area from %d to %d",
			&colS, &colE) != 2
	        &&!sscanf(buf, "Color rows: %d", &rowNum))
	        {
		 sprintf(XDim->error, XDERR_SYNTAX_S, buf);
		 status =ERROR;
		 break;
	        }
	      i++;
	     }
	  }
       row = 0;
       wanted = 0;
       for(i=0; i<rowNum;)
          {
	   if(*buf != '#')
	     {
              if(fgets(buf, MAX_NAME_LENGTH, in)
                 && sscanf(buf,"%d. START:  R: %hu   G: %hu   B: %hu",
        	    &dummy,&sR, &sG, &sB)==4
        	 && fgets(buf, MAX_NAME_LENGTH, in)
        	 && sscanf(buf,"   END:  R: %hu   G: %hu   B: %hu   Depth: %d\n",
        	    &eR, &eG, &eB, &depth))
        	{
		 if(depth >0)
		   {
        	    row++;
        	    wanted += depth;
        	    (XDC->colors)[i] = depth;
       		    (XDC->beginR)[i] = sR;
       		    (XDC->beginG)[i] = sG;
       		    (XDC->beginB)[i] = sB;
       		    (XDC->endR)[i] = eR;
       		    (XDC->endG)[i] = eG;
       		    (XDC->endB)[i] = eB;
		    i++;
       		   }
        	}
              else
        	{
        	 /* syntax error */
		 sprintf(XDim->error, XDERR_SYNTAX_S, buf);
		 status = ERROR;
		 break;
        	}
	     }
          }
        if(row > 0)
          {
           XDC->rowNum = row;
           for(i=0; i< XDC->wantedColors; i++)
              if((HWI->colStat)[i])
                (HWI->colStat)[i] = HW_COLORCLAIMED;
           XDC->wantedColors = wanted;
          }
       if(colS >= XDC->wantedColors)
          XDC->colS = 0;
       else
          XDC->colS = colS;
       if(colE >= XDC->wantedColors)
          XDC->colE = XDC->wantedColors -1;
       else
          XDC->colE = colE;
       if(fore < XDC->wantedColors)
         XDC->fore = fore;
       else
         XDC->fore = 0;
       if(fore < XDC->wantedColors)
         XDC->back = back;
       else
         XDC->back = 1;
       if(textFore < XDC->wantedColors)
         XDC->textFore = textFore;
       else
         XDC->textFore = fore;
       if(axisFore < XDC->wantedColors)
         XDC->axisFore = axisFore;
       else
         XDC->axisFore = fore;
       if(axisBX < XDC->wantedColors)
         XDC->axisBackX = axisBX;
       else
         XDC->axisBackX = back;
       if(axisBY < XDC->wantedColors)
         XDC->axisBackY = axisBY;
       else
         XDC->axisBackY = back;
       if(axisBZ < XDC->wantedColors)
         XDC->axisBackZ = axisBZ;
       else
         XDC->axisBackZ = back;
       XDC->tInterval = tInterval;
       if((redfunc = XT_GetNamedRedraw(FRedraw, XDim)) != NULL)
	XDC->FastRedraw= redfunc;
       if((redfunc = XT_GetNamedRedraw(NRedraw, XDim)) != NULL)
	XDC->NormalRedraw = redfunc;
       XDC->interCol = interCol;
       fclose(in);
       return(status);
      }
    sprintf(XDim->error, XDERR_OPEN_READ_S, XDC->pref);
    return(ERROR);
   }

/*
 * Save a deco file
 */

int FI_SaveDeco(XDimInfo *XDim)
   {
    FILE *out;
    ViewportInfo *VWI;
    int i;
    myPixel *deco;

    if(XDim->port >=0)
      {
       VWI = &((XDim->Views)[XDim->port]);
       out = fopen(VWI->dfile, "w");
       if(out)
	 {
	  fprintf(out, "XDim deco file:\n");
	  fprintf(out, "Decoding entries: %d\n", MAX_DECO_NUMBER);
          deco = VWI->deco;
	  for(i=0; i<MAX_DECO_NUMBER; i++)
             {
              fprintf(out,"%d. %d\n",(i+1),*deco++);
             }
	  fclose(out);
	  return(OK);
	 }
       else
	 {
	  sprintf(XDim->error, XDERR_OPEN_WRITE_S, VWI->dfile);
	  return(ERROR);
	 }
      }
    strcpy(XDim->error, XDERR_NO_VIEW);
    return(ERROR);
   }

/*
 * Load a deco file
 */

int FI_LoadDeco(XDimInfo *XDim)
   {
    FILE *in;
    ViewportInfo *VWI;
    XDColorInfo *XDC;
    char buf[MAX_NAME_LENGTH], *status;
    int i, num, Sum, dummy;
    unsigned long buffer[MAX_DECO_NUMBER], *val, max;
    myPixel *deco;
    double fact;

    if(XDim->port >=0)
      {
       VWI = &((XDim->Views)[XDim->port]);
       in = fopen(VWI->dfile, "r");
       if(in)
	 {
	  fgets(buf, MAX_NAME_LENGTH, in);
	  if(strncmp(buf, "XDim deco file:\n", MAX_NAME_LENGTH))
	    {
	     strcpy(XDim->error, XDERR_NO_DEC);
	     fclose(in);
             return(ERROR);
            }
	  num = -1;
	  while(fgets(buf, MAX_NAME_LENGTH, in))
             {
	      if(*buf != '#')
		{
		 if(sscanf(buf, "Decoding entries: %d", &num))
		   break;
		 else
	           {
		    sprintf(XDim->error, XDERR_SYNTAX_S, buf);
		    return(ERROR);
	           }
	        }
	     }
	  if(num < 0)
	    {
	     sprintf(XDim->error, XDERR_NUM_DEC);
	     return(ERROR);
	    }
	  if(num <= MAX_DECO_NUMBER)
	    {
	     Sum = MAX_DECO_NUMBER-1;
	     val = buffer;
	     for(i=0; i<MAX_DECO_NUMBER;i++)
	        {
	         Sum += num;
	         if(Sum >= MAX_DECO_NUMBER)
	           {
	            Sum -= MAX_DECO_NUMBER;
	            while((status=fgets(buf, MAX_NAME_LENGTH, in)))
	              {
	               if(*buf != '#')
	                 {
		          if(sscanf(buf,"%d. %ld\n",&dummy,&max)==2)
		             break;
		          else
		            {
		             sprintf(XDim->error, XDERR_SYNTAX_S, buf);
		             return(ERROR);
		            } 
		         }            
	              }
	            if(!status)
	              {
		       strcpy(XDim->error, XDERR_EOF);
		       return(ERROR);
	              }
	           }
	         *val++ = max;
	        }
	    }
	  else
	    {
	     /* num > MAX_DECO_NUMBER */
	     Sum = num-1;
	     val = buffer;
	     for(i=0; i<num;)
	        {
	         if(!fgets(buf, MAX_NAME_LENGTH, in))
	           {
		    strcpy(XDim->error, XDERR_EOF);
		    return(ERROR);
	           }
	         if(*buf != '#')
	           {
	            Sum += MAX_DECO_NUMBER;
	            if(Sum >= num)
	              {
	               Sum -= num;
		       if(sscanf(buf,"%d. %ld\n",&dummy,val)==2)
		         {
		          i++;
		          val++;
		         }
		       else
		         {
		          sprintf(XDim->error, XDERR_SYNTAX_S, buf);
		          return(ERROR);
		         } 
	              }
	           }
	        }
	    }
          val = buffer;
          deco = VWI->deco;
          VWI->changed |= TRANS_AND_CHANGE;
	  XDC = &(XDim->colors);
	  max = *val;
          for(i=1; i<MAX_DECO_NUMBER; i++)
             {
              if(*val > max)
                max = *val;
              val++;
             }
          if(max > (XDC->wantedColors-1))
	    fact = ((double) (XDC->wantedColors-1))/max;
	  else
	    fact =1.0;
	  val = buffer;
	  for(i=0; i<MAX_DECO_NUMBER; i++)
	     *deco++ = (myPixel)(fact* *val++);
	  return(OK);
	 }
       else
	 {
	  sprintf(XDim->error, XDERR_OPEN_READ_S, VWI->dfile);
	  return(ERROR);
	 }
      }
    strcpy(XDim->error, XDERR_NO_VIEW);
    return(ERROR);
   }

/*
 * load a map file
 */

int FI_LoadMap(char *name, XDimInfo *XDim)
   {
    FILE *in;
    XDColorInfo *XDC;
    XDColMap *cm, *nm;
    char buf[MAX_NAME_LENGTH], up;
    int maxMapped, mapped, r, g, b, num;

    if(XDim->port >=0)
      {
       in = fopen(name, "r");
       if(in)
	 {
	  XDC = &(XDim->colors);
	  mapped = 0;
	  maxMapped = 0;
	  cm = NULL;
	  while(fgets(buf, MAX_NAME_LENGTH, in))
	    {
	     if(mapped == maxMapped)
	       {
	        /* allocate more memory */
	        maxMapped += 64;
	        if(maxMapped > MAX_MAPPED)
	          break;
	        nm = realloc(cm, maxMapped*sizeof(XDColMap));
	        if(nm == NULL)
	          break;
	        else
	          cm = nm;
	       }
	     up = 'u';
	     r = 0;
	     g = 0;
	     b = 0;
	     num = sscanf(buf, "r: %d g: %d b: %d %c", &r, &g, &b, &up);
	     if(num >= 3)
	       {
	        nm = &(cm[mapped++]);
	        nm->red = r;
	        nm->green = g;
	        nm->blue = b;
	        if(up == 'd')
	          nm->botUp = True;
	        else
	          nm->botUp = False;
	       }
	    }
	  if(mapped > 0)
	    {
	     XDC->colMap = cm;
	     XDC->maxMapped = mapped;
	     return(OK);
	    }
	  else
	    strcpy(XDim->error, XDERR_EOF);
	 }
       else
         sprintf(XDim->error, XDERR_OPEN_READ_S, name);
      }
    else
      strcpy(XDim->error, XDERR_NO_VIEW);
    return(ERROR);
   }

/*
 * Save a data file
 */

int FI_SaveData(XDimInfo *XDim)
   {
    FILE *out;
    ViewportInfo *VWI;
    char *link2File;
    unsigned char *cp;
    int i, j, Size, Num;
    double *dp;

    if(XDim->port >=0)
      {
       VWI = &((XDim->Views)[XDim->port]);
       if(VWI->actData == NULL && VWI->charData == NULL)
         {
	  strcpy(XDim->error, XDERR_NO_DATA);
	  return(ERROR);
         }
       out = fopen(VWI->fileName, "w");
       if(out)
	 {
	  if(VWI->dataType)
	    fprintf(out, "XDim data file: Type %d\n", VWI->dataType);
	  else
	    fprintf(out, "XDim data file:\n");
	  fprintf(out, "Max X: %d\n", VWI->Width);
	  fprintf(out, "Max Y: %d\n", VWI->Height);
	  fprintf(out, "Dim X: %s\n", VWI->DimX);
	  fprintf(out, "Dim Y: %s\n", VWI->DimY);
	  fprintf(out, "Dim Z: %s\n", VWI->DimZ);
	  fprintf(out, "Offset X: %e\n", VWI->OffX);
	  fprintf(out, "Form Factor X: %e\n", VWI->FactX);
	  fprintf(out, "Offset Y: %e\n", VWI->OffY);
	  fprintf(out, "Form Factor Y: %e\n", VWI->FactY);
	  if(XDim->useLinks && VWI->links >0)
	    {
	     fprintf(out,"links:\n");
	     Num = 1;
	     for(i=0;i<VWI->links;i++)
		{
		 link2File = ((XDim->Views)[(VWI->Link2)[i]]).fileName;
		 if(*link2File != '\0')
	            fprintf(out,"link %d: %s to %s in %s\n", Num++,
			(VWI->LinkItself)[i], (VWI->LinkName)[i], link2File);
		}
	    }
	  fprintf(out, "Data:\n");
	  Size = VWI->Width * VWI->Height;
	  switch(VWI->dataType)
	     {
              case MY_UCHAR:
	      cp = VWI->charData;
	      for(i=0; i<Size; i++)
        	  fprintf(out,"%cu\n",*cp++);
	      break;
	      /*
	       * Triangle = 3 Points with 3 coors + 1 color
	       */
	      case MY_TRIANGLE:
	      dp = VWI->actData;
	      for(i=0; i< VWI->Width; i++)
	         {
	          for(j=0; j < VWI->Height; j++)
	              fprintf(out,"(%e %e %e) ", *dp++, *dp++, *dp++);
	          /* color */
	          fprintf(out,"%e\n", *dp++);
	         }
	      break;
	      default:
	      dp = VWI->actData;
	      for(i=0; i<Size; i++)
        	  fprintf(out,"%e\n",*dp++);
              break;
             }
	  fclose(out);
	  return(OK);
	 }
       else
	 {
	  sprintf(XDim->error, XDERR_OPEN_WRITE_S, VWI->fileName);
	  return(ERROR);
	 }
      }
    strcpy(XDim->error, XDERR_NO_VIEW);
    return(ERROR);
   }

/*
 * Load a data file
 */

int FI_LoadData(XDLinks *hLinks, XDimInfo *XDim)
   {
    FILE *in;
    ViewportInfo *VWI;
    char buf[MAX_NAME_LENGTH+2*MAX_LINK_LENGTH+16], *ip;
    unsigned char *cp, cin;
    int Size, i, j, NewWidth, NewHeight, dataType, num, dum;
    char DimX[MAX_DIM_LENGTH], DimY[MAX_DIM_LENGTH], DimZ[MAX_DIM_LENGTH];
    double val, vals[20], *bp, *dp, FactX, FactY,OffX,OffY;

    if(XDim->port >=0)
      {
       VWI = &((XDim->Views)[XDim->port]);
       if(VWI->edit && VWI->port == XDim->aView.edPort)
         XDim->aView.edPort = -1;
       in = fopen(VWI->fileName, "r");
       if(in)
	 {
	  fgets(buf, MAX_NAME_LENGTH, in);
	  if(strncmp(buf, "XDim data file\n", 9))
	    {
	     strcpy(XDim->error, XDERR_NO_XDIM_DATA);
	     fclose(in);
             return(ERROR);
            }
          dataType = MY_DOUBLE;
          sscanf(buf,"XDim data file: Type %d", &dataType);
	  i= 0;
	  NewWidth = 2;
	  NewHeight = 2;
	  DimX[0] = '\0';
	  DimY[0] = '\0';
	  DimZ[0] = '\0';
	  FactX = 1.0;
	  FactY = 1.0;
	  OffX = 0.0;
	  OffY = 0.0;
          VWI->links = 0;
          if(hLinks != NULL)
            {
             hLinks->links = 0;
	     *((hLinks->hLinks)[0]).Itself = '\0';
	     *((hLinks->hLinks)[0]).LinkName = '\0';
	     *((hLinks->hLinks)[0]).Link2File = '\0';
	    }
	  while(i<NUM_DATA && fgets(buf, MAX_NAME_LENGTH, in))
             {
	      if(*buf != '#')
		{
		 if(!sscanf(buf,"Max X: %d" , &NewWidth)
	           &&!sscanf(buf, "Max Y: %d", &NewHeight)
	           &&!sscanf(buf, "Dim X: %s", DimX)
	           &&!sscanf(buf, "Dim Y: %s", DimY)
	           &&!sscanf(buf, "Dim Z: %s", DimZ)
	           &&!sscanf(buf, "Offset X: %le", &OffX)
	           &&!sscanf(buf, "Form Factor X: %le", &FactX)
	           &&!sscanf(buf, "Offset Y: %le", &OffY)
	           &&!sscanf(buf, "Form Factor Y: %le", &FactY))
	           {
		    sprintf(XDim->error, XDERR_SYNTAX_S, buf);
		    return(ERROR);
	           }
		 i++;
		}
	     }
	  if(i!=NUM_DATA)
	    {
	     strcpy(XDim->error, XDERR_MORE_INFO);
	     return(ERROR);
	    }
	  if(CO_NewData(NewWidth, NewHeight, dataType, XDim))
	    return(ERROR);
	
	  Size = NewWidth*NewHeight;
	  i = 0;
	  strcpy(VWI->DimX, DimX);
	  strcpy(VWI->DimY, DimY);
	  strcpy(VWI->DimZ, DimZ);
	  if(dataType < MY_TRIANGLE)
	    {
	     if(VWI->zBuf)
	       {
	        free(VWI->zBuf);
	        VWI->zBuf = NULL;
	       }
	     VWI->FactX = FactX;
	     VWI->FactY = FactY;
	     VWI->OffX = OffX;
	     VWI->OffY = OffY;
	     if(NewWidth > NewHeight)
	       {
		if(VWI->gNum > NewHeight)
		  VWI->gNum = NewHeight;
		if(VWI->gStep > NewHeight)
		  VWI->gStep = NewHeight >> 2;
	       }
	     else
	       {
		if(VWI->gNum > NewWidth)
		  VWI->gNum = NewWidth;
		if(VWI->gStep > NewWidth)
		  VWI->gStep = NewWidth >> 2;
	       }
	    }
	  if(VWI->gStep == 0)
	    VWI->gStep = 1;
	  while(fgets(buf, MAX_NAME_LENGTH, in) != NULL &&
		strncmp(buf, "Data:", 4))
	      {
               if(hLinks != NULL && hLinks->links < MAX_VIEWPORT_NUMBER &&
		sscanf(buf, "link %d: %s to %s in %s", &dum,
			((hLinks->hLinks)[hLinks->links]).Itself,
			((hLinks->hLinks)[hLinks->links]).LinkName,
			((hLinks->hLinks)[hLinks->links]).Link2File))
		 {
		  (hLinks->links)++;
		  if(hLinks->links < MAX_VIEWPORT_NUMBER)
		    {
		     *((hLinks->hLinks)[hLinks->links]).Itself = '\0';
		     *((hLinks->hLinks)[hLinks->links]).LinkName = '\0';
		     *((hLinks->hLinks)[hLinks->links]).Link2File = '\0';
		    }
		 }
	      }

	  /* Now read data in according to the type */

	  switch (dataType)
             {
              case MY_UCHAR:
	      cp = VWI->charData;
              while(i < Size && fgets(buf, MAX_NAME_LENGTH, in) != NULL)
        	  {
        	   if(sscanf(buf, "%cu", &cin))
                     {
                      i++;
                      *cp++ = cin;
                     }
        	  }
              break;
              case MY_TRIANGLE:
              /* 3 points * 3 coors + color */
              Size = VWI->Width;
	      dp = VWI->actData;
	      num = 3*VWI->Height;
              while(i < Size && fgets(buf, MAX_NAME_LENGTH, in) != NULL)
        	  {
		   bp = vals;
		   ip = buf;
	           for(j=0; j < VWI->Height; j++)
	              {
	               while(*ip != '\0' && '('!=*ip++);
	               if(sscanf(ip,"%le %le %le", bp, bp+1, bp+2) != 3)
	                break;
	               bp += 3;
	              }
	           /* color */
	           if(j == VWI->Height && *ip != '\0')
	             {
	              while(*ip != '\0' && ')'!=*ip++);
	              if((dum = sscanf(ip,"%le %le %le\n", bp, bp+1, bp+2))>0)
	                {
	                 i++;
	                 bp = vals;
	                 for(j =0; j< num; j++)
	                    *dp++ = *bp++;
	                 if(dum < COLOR_3D)
	                   {
	                    for(j=0; j < COLOR_3D; j++)
	                        *dp++ = *bp;
	                   }
	                 else
	                   {
	                    for(j=0; j < COLOR_3D; j++)
	                        *dp++ = *bp++;
	                   }
	                }
	             }
	          }
	      break;
              default:
	      dp = VWI->actData;
              while(i < Size && fgets(buf, MAX_NAME_LENGTH, in) != NULL)
        	  {
        	   if(sscanf(buf, "%le", &val))
                     {
                      i++;
                      *dp++ = val;
                     }
        	  }
	      break;
	     }
	  if(i!=Size)
	    {
	     strcpy(XDim->error, XDERR_EOF);
	     free(VWI->actData);
	     VWI->actData = NULL;
	     return(ERROR);
	    }
	  else
	    return(OK);
	 }
       else
	 {
	  sprintf(XDim->error, XDERR_OPEN_READ_S, VWI->fileName);
	  return(ERROR);
	 }
      }
    strcpy(XDim->error, XDERR_NO_VIEW);
    return(ERROR);
   }

/*
 * Process buffer for ASCII import
 */

void FI_ProcASCIIBuf(char *s, int max)
   {
    char ch, *sp;
    int i;
    
    sp = s;
    for(i= 0; i<max; i++)
       {
        ch = *sp++;
        if(ch == '\0')
          break;
        if(ch == ',')
          {
           if(isdigit(*sp) || (i>0 && isdigit(*(sp-2))))
              *(sp-1) = '.';
          }
       }
    
   }

/*
 * Get next number from ASCII buffer
 */

char *FI_GetNextNum(char *s, double *val)
   {
    char ch, *sp;
    
    *val = strtod(s, &sp);
    if(sp==s)
      sp = NULL;
    if(sp !=NULL)
      {
       while((ch = *sp)!='\0')
          {
           if(ch == ' ' || ch == ',' || ch == ';' || ch == '\t'
		|| ch == ':')
	      break;
	   sp++;             
          }
      }
    return(sp);
   }

/*
 * Import an ASCII data file
 */

int FI_ImportAsciiData(XDimInfo *XDim)
   {
    FILE *in;
    ViewportInfo *VWI;
    XDColorInfo *XDC;
    char *sp, buf[MAX_NAME_LENGTH<<6];
    int Size, Length, i, j, NewWidth, NewHeight, dataType;
    double val, *dp, Repl;

    if(XDim->port >=0)
      {
       VWI = &((XDim->Views)[XDim->port]);
       XDC = &(XDim->colors);
       in = fopen(VWI->fileName, "r");
       if(in)
	 {
	  Length = MAX_NAME_LENGTH<<6;
	  dataType = MY_DOUBLE;
	  NewWidth = 0;
	  NewHeight = 0;
	  strcpy(VWI->DimX, "XXX");
	  strcpy(VWI->DimY, "XXX");
	  strcpy(VWI->DimZ, "XXX");
	  VWI->FactX = 1.0;
	  VWI->FactY = 1.0;
	  VWI->OffX = 0.0;
	  VWI->OffY = 0.0;
          VWI->links = 0;
          while(fgets(buf, Length, in) != NULL)
              {
               if(*buf != '#' && *buf !='"')
                 {
                  if(XDC->GerNot == True)
                    FI_ProcASCIIBuf(buf, Length);
                  sp = buf;
                  if(NewWidth)
                    {
                     if(FI_GetNextNum(sp, &val) != NULL)
                       NewHeight++;
                    }
                  else
                    {
                     while((sp=FI_GetNextNum(sp, &val)) != NULL)
                        NewWidth++;
                     if(NewWidth != 0)
                        NewHeight++;
                    }
                 }
	      }
	  Size = NewWidth*NewHeight;
	  if(Size == 0)
	    {
	     strcpy(XDim->error, XDERR_FORMAT);
	     return(ERROR);
	    }
	  if(CO_NewData(NewWidth, NewHeight, dataType, XDim))
	    return(ERROR);
	  Repl = XDC->Repl;
	  dp = VWI->actData;
          rewind(in);
          for(i= 0; i< NewHeight; i++)
             {
              fgets(buf, Length, in);
              if(*buf != '#' && *buf != '"')
                {
                 sp = buf;
                 if(XDC->GerNot == True)
                    FI_ProcASCIIBuf(buf, Length);
                 if((sp=FI_GetNextNum(sp, &val)) != NULL)
                   {
                    *dp++ = val;
                    for(j=1; j< NewWidth; j++)
                       {
                	if((sp=FI_GetNextNum(sp, &val)) == NULL)
                	  val = Repl;
                	*dp++ = val;
                       }
                   }
                 else
                   i--;
                }
              else
                i--;
             }
	  if(NewWidth == 1)
	    {
	     /* unstructured data */
	     NewWidth = (int) sqrt((double)NewHeight);
	     for(i=NewWidth; i >0; i--)
		{
		 j = NewHeight/i;
		 if(i*j==NewHeight)
                   break;
		}
	     VWI->Width = i;
	     VWI->Height = j;
	     strcpy( XDim->mess, XMESS_UNSTRUCT);
	     WD_ViewportMessage(VWI, XDim);
	    }
	  if(NewWidth > NewHeight)
	    {
	     if(VWI->gNum > NewHeight)
	       VWI->gNum = NewHeight;
	     if(VWI->gStep > NewHeight)
	       VWI->gStep = NewHeight >> 2;
	    }
	  else
	    {
	     if(VWI->gNum > NewWidth)
	       VWI->gNum = NewWidth;
	     if(VWI->gStep > NewWidth)
	       VWI->gStep = NewWidth >> 2;
	    }
	  if(VWI->gStep == 0)
	    VWI->gStep = 1;
	  return(OK);
	 }
       else
	 {
	  sprintf(XDim->error, XDERR_OPEN_READ_S, VWI->fileName);
	  return(ERROR);
	 }
      }
    strcpy(XDim->error, XDERR_NO_VIEW);
    return(ERROR);
   }


/*
 * Export an ASCII data file
 */

int FI_AsciiExportData(char *FName, XDimInfo *XDim)
   {
    FILE *out;
    ViewportInfo *VWI;
    XDColorInfo *XDC;
    int i, j;
    double *dp;
    char buf[MAX_NAME_LENGTH], *sp;

    if(XDim->port >=0)
      {
       VWI = &((XDim->Views)[XDim->port]);
       XDC = &(XDim->colors);
       if(VWI->actData == NULL)
         {
	  strcpy(XDim->error, XDERR_NO_DATA);
	  return(ERROR);
         }
       out = fopen(FName, "w");
       if(out)
	 {
	  fprintf(out, "\"Max X: %d\"\n", VWI->Width);
	  fprintf(out, "\"Max Y: %d\"\n", VWI->Height);
	  fprintf(out, "\"Dim X: %s\"\n", VWI->DimX);
	  fprintf(out, "\"Dim Y: %s\"\n", VWI->DimY);
	  fprintf(out, "\"Dim Z: %s\"\n", VWI->DimZ);
	  fprintf(out, "\"Offset X: %e\"\n", VWI->OffX);
	  fprintf(out, "\"Form Factor X: %e\"\n", VWI->FactX);
	  fprintf(out, "\"Offset Y: %e\"\n", VWI->OffY);
	  fprintf(out, "\"Form Factor Y: %e\"\n", VWI->FactY);
	  fprintf(out, "\"Data:\"\n");
	  dp = VWI->actData;
          if(XDC->GerNot)
            {
	     for(i=0; i<VWI->Height; i++)
        	{
        	 for(j=0; j < VWI->Width; j++)
                    {
                     sprintf(buf, "\t%e", *dp++);
                     sp = strchr(buf, '.');
                     if(sp != NULL)
                       *sp = ',';
                     fputs(buf, out);
                    }
                  fputc('\n', out);
                 }
             }
           else
             {
	      for(i=0; i<VWI->Height; i++)
        	 {
        	  for(j=0; j < VWI->Width; j++)
                    {
                     fprintf(out,"\t%e",*dp++);
                    }
                  fputc('\n', out);
                 }
             }
	  fclose(out);
	  return(OK);
	 }
       else
	 {
	  sprintf(XDim->error, XDERR_OPEN_WRITE_S, FName);
	  return(ERROR);
	 }
      }
    strcpy(XDim->error, XDERR_NO_VIEW);
    return(ERROR);
   }

/*
 * Return a Token for the 3D data type
 */

int FI_FindToken(char *start, int len)
   {
    int i, ret, num;
    char ch, *buf;
    double val;
    
    ret = -1;
    i = 0;
    buf = start;
    do
      {
       ch = *buf++;
       i++;
      }
    while(ch != '\0' && ch == ' ' && ch == '\n' && ch == '\t' && i < len);
    switch(ch)
      {
       case '"':
       case '#':
       case '\0':
       ret = XD_COMMENT;
       break;
       case 'p':
       case 'P':
       ret = XD_POLY;
       break;
       case 'l':
       case 'L':
       ret = XD_LINE;
       break;
       case 'e':
       case 'E':
       ret = XD_ELLIPS;
       break;
       case 'o':
       case 'O':
       ret = XD_ONE_IND;
       break;
       case 'r':
       case 'R':
       if(*buf == 'g' || *buf == 'G')
         ret = XD_RGB;
       else
         ret = XD_ROT;
       break;
       case 'v':
       case 'V':
       ret = XD_PER_VALUE;
       break;
       case 'z':
       case 'Z':
       ret = XD_ZCOL;
       break; 
      }
    if(ret < 0)
      {
       /*
        * test if polygon or 3D Data
        */
       num = 0;
       buf -= 1;
       while((buf = O3_FindNumber(&val, buf, start, True)) != NULL)
         num++;
       if(num == 3 || num == 4)
         ret = XD_3D_DATA;
       else if(num > 8)
         ret = XD_POLY;
      }
    return(ret);
   }
/*
 * Set the color import mode
 */

void FI_SetColorIMode(int *opt, int val)
   {
    int nopt, mask;

    nopt = *opt;
    mask = XD_3D_RGB | XD_3D_ZCOL | XD_3D_PER_VALUE | XD_3D_ONE_IND;
    nopt &= (~mask);
    nopt |= val;
    *opt = nopt;
   }

/*
 * Import ASCII 3D
 * Warning:
 * This is only a convenience format allowing the
 * simple creation of data
 * Especially it's impossible to export this format
 * Please use VRML/OpenGL for this
 */

int FI_Import3DAscii(XDimInfo *XDim)
   {
    FILE *in;
    ViewportInfo *VWI;
    XDColorInfo *XDC;
    XDHardwareInfo *HWI;
    char buf[MAX_NAME_LENGTH<<6];
    unsigned char red[MAX_COL_NUMBER], green[MAX_COL_NUMBER];
    unsigned char blue[MAX_COL_NUMBER];
    unsigned short dum;
    int Size, freeP, Length, i, NewWidth, NewHeight, dataType, status;
    int num_colors, col_end, ind1, ind2, rows;
    double *data, *dp, *dp1;

    if(XDim->port >=0)
      {
       VWI = &((XDim->Views)[XDim->port]);
       XDC = &(XDim->colors);
       HWI = &(XDim->hwInfo);
       in = fopen(VWI->fileName, "r");
       if(in)
	 {
	  FI_SetColorIMode(&(XDC->opt3d), XD_3D_ONE_IND);
	  Length = MAX_NAME_LENGTH<<6;
	  dataType = MY_TRIANGLE;
	  NewWidth = 0;
	  NewHeight = 3;
	  num_colors = 0;
	  strcpy(VWI->DimX, "XXX");
	  strcpy(VWI->DimY, "XXX");
	  strcpy(VWI->DimZ, "XXX");
	  VWI->FactX = 1.0;
	  VWI->FactY = 1.0;
	  VWI->OffX = 0.0;
	  VWI->OffY = 0.0;
          VWI->links = 0;
          status = OK;
          data = NULL;
          freeP = 0;
          while(fgets(buf, Length, in) != NULL)
              {
               switch(FI_FindToken(buf, Length))
                  {
                   case XD_POLY:
                   if(O3_InsertPoly(buf, Length, &data, &NewWidth,
			&freeP, &num_colors, red, green, blue, XDC, XDim))
                      status = ERROR;
                   break;
                   case XD_LINE:
		   if(O3_InsertLines(buf, Length, &data, &NewWidth,
			&freeP, &num_colors, red, green, blue, XDC, XDim))
                      status = ERROR;
		   break;
                   case XD_ELLIPS:
		   if(O3_InsertEll(buf, Length, &data, &NewWidth, &freeP,
			&num_colors, red, green, blue, in, XDC, XDim))
		      status= ERROR;
                   break;
                   case XD_3D_DATA:
		   if(O3_Insert3D(buf, Length, &data, &NewWidth, &freeP,
			&num_colors, in, XDC, XDim))
		      status = ERROR;
                   break;
                   case XD_RGB:
                   if(NewWidth)
                     {
                      status = ERROR;
                      strcpy(XDim->error, XDERR_RGB);
                      continue;
                     }
                   else
                      FI_SetColorIMode(&(XDC->opt3d) , XD_3D_RGB);
                   break;
                   case XD_ZCOL:
                      FI_SetColorIMode(&(XDC->opt3d), XD_3D_ZCOL);
                   break;
                   case XD_PER_VALUE:
                      FI_SetColorIMode(&(XDC->opt3d), XD_3D_PER_VALUE);
                   break;
		   case XD_ONE_IND:
		      FI_SetColorIMode(&(XDC->opt3d), XD_3D_ONE_IND);
		   break;
                   default:
                   status = ERROR;
		   sprintf(XDim->error, XDERR_SYNTAX_S, buf);
                   case XD_COMMENT:
                   break;
                  }
               if(status) break;
	      }
	  if(status)
	    {
	     NewWidth = 0;
	     if(data)
               {
		free(data);
		data = NULL;
               }
            }
          if(data != NULL)
            {
	     if(CO_NewData(NewWidth, NewHeight, dataType, XDim))
		status = ERROR;
	     else
	       {
	        dp = data;
		dp1 = VWI->actData;
		Size = NewWidth * (3*NewHeight+COLOR_3D);
		for(i=0; i<Size; i++)
		   *dp1++ = *dp++;
	       }
             free(data);
            }
	  if(XDC->opt3d & XD_3D_RGB)
	    {
	     rows = 0;
	     for(i=0; i<num_colors >>1; i++)
        	{
        	 (XDC->colors)[i] = 2;
        	 ind1 = i<<1;
        	 ind2 = (i<<1)+1;
		 dum =  red[ind1];
       		 (XDC->beginR)[i] = (dum<<8)|dum;
		 dum =  green[ind1];
       		 (XDC->beginG)[i] = (dum<<8)|dum;
		 dum = blue[ind1];
       		 (XDC->beginB)[i] = (dum<<8)|dum;
		 dum =  red[ind2];
       		 (XDC->endR)[i] = (dum<<8)|dum;
		 dum =  green[ind2];
       		 (XDC->endG)[i] = (dum<<8)|dum;
		 dum =  blue[ind2];
       		 (XDC->endB)[i] = (dum<<8)|dum;
       		 rows++;
        	}
             if(num_colors&1)
               {
                ind1 = num_colors-1;
                ind2 = num_colors-1;
        	(XDC->colors)[rows] = 1;
		dum =  red[ind1];
       		(XDC->beginR)[rows] = (dum<<8)|dum;
		dum =  green[ind1];
       		(XDC->beginG)[rows] = (dum<<8)|dum;
		dum = blue[ind1];
       		(XDC->beginB)[rows] = (dum<<8)|dum;
		dum =  red[ind2];
       		(XDC->endR)[rows] = (dum<<8)|dum;
		dum =  green[ind2];
       		(XDC->endG)[rows] = (dum<<8)|dum;
		dum =  blue[ind2];
       		(XDC->endB)[rows] = (dum<<8)|dum;
       		rows++;
               }
             col_end = num_colors;
             if(num_colors < 4)
               {
                num_colors += 4;
        	(XDC->colors)[rows] = 4;
       		(XDC->beginR)[rows] = 0;
       		(XDC->beginG)[rows] = 0;
       		(XDC->beginB)[rows] = 0;
       		(XDC->endR)[rows] = USHRT_MAX;
       		(XDC->endG)[rows] = USHRT_MAX;
       		(XDC->endB)[rows] = USHRT_MAX;
       		rows++;
               }
             XDC->rowNum = rows;
	     GR_InitColorImport( num_colors, col_end, XDC, HWI);
	    }
	  return(status);
	 }
       else
	 {
	  sprintf(XDim->error, XDERR_OPEN_READ_S, VWI->fileName);
	  return(ERROR);
	 }
      }
    strcpy(XDim->error, XDERR_NO_VIEW);
    return(ERROR);
   }

/*
 * Export a GIF data picture
 */

int FI_GifExportPic(char *FName, XDimInfo *XDim)
   {
    ViewportInfo *view;
    XDColorInfo *XDC;
    XDHardwareInfo *HWI;
    int *red, *green, *blue, *r, *g, *b, Size, bpp, intrlce, back, i;

    if(XDim->port >=0)
      {
       view = &((XDim->Views)[XDim->port]);
       XDC = &(XDim->colors);
       HWI = &(XDim->hwInfo);
       if(view->pict == NULL)
         {
	  strcpy(XDim->error, XDERR_NO_IMAGE);
	  return(ERROR);
         }
       bpp = (sizeof(myPixel)) << CHAR2BITS;
       for(i=1; i <= bpp; i++)
          {
           Size = 1 << i;
           if(Size >= XDC->wantedColors)
             break;
          }
       if(Size > 256)
         {
	  strcpy(XDim->error, XDERR_GIF_COL);
	  return(ERROR);
         }
       bpp = i;
       Size *= sizeof(int);
       red = (int*) malloc(Size);
       green = (int*) malloc(Size);
       blue = (int*) malloc(Size);
       if(red == NULL || green == NULL || blue == NULL)
         {
          free(red);
          free(green);
          free(blue);
	  strcpy(XDim->error, XDERR_NO_MEM);
	  return(ERROR);
         }
       r = red;
       g = green;
       b = blue;
       for(i=0;i<XDC->wantedColors; i++)
          {
           *r++ = (int) (HWI->red)[i];
           *g++ = (int) (HWI->green)[i];
           *b++ = (int) (HWI->blue)[i];
          }
       intrlce = False;
       back = view->back;
       GIFEncode(FName, view->pwidth, view->pheight, intrlce, back,
		bpp, red, green, blue, GR_ReadPixel);
       return(OK);
      }
    strcpy(XDim->error, XDERR_NO_VIEW);
    return(ERROR);
   }


/* These static variables are needed by GIFerror */

static jmp_buf setjmp_buffer;	/* for return to caller */

static unsigned char *FI_GifPic;
static int FI_GifPWidth;

/*
 * Report GIF import error
 */

void FI_GIFerror(char *s1, char *s2)
  {
   extern XDimInfo XDim;

   if(s2 != NULL)
     sprintf(XDim.error, s1, s2);
   else
     strcpy(XDim.error, s1);
   WD_XDimError(&XDim, NULL);
   GIFfreeAll();
   longjmp(setjmp_buffer, 1);	/* return control to outer routine */
  }

/*
 * Allocate memory to store values
 */

int FI_MemAlloc(int width, int height)
  {
   int ret;
   ViewportInfo *view;
   extern XDimInfo XDim;

   if(CO_NewData(width, height, MY_UCHAR, &XDim))
     ret = False;
   else
     {
      ret = True;
      view = &((XDim.Views)[XDim.port]);
      FI_GifPic = view->charData;
      FI_GifPWidth = view->Width;
     }
   return(ret);
  }

/*
 * Write the value at x,y
 */

int FI_WriteValue(int x, int y, int index)
  {
   *(FI_GifPic+y*FI_GifPWidth+x) = (unsigned char)index;
   return(TRUE);
  }

/*
 * Import a gif picture
 */

int FI_ImportGifPic(XDimInfo *XDim)
   {
    ViewportInfo *view;
    XDColorInfo *XDC;
    XDHardwareInfo *HWI;
    int i, rows, num_colors, ind1, ind2;
    unsigned short dum;
    unsigned char red[256], green[256], blue[256];

    if(XDim->port >=0)
      {
       view = &((XDim->Views)[XDim->port]);
       XDC = &(XDim->colors);
       HWI = &(XDim->hwInfo);
       if (setjmp(setjmp_buffer))
          {
  /* If we get here, the GIF code has signaled an error. */
	   return(ERROR);
	  }
       GIFdecode(view->fileName, red, green, blue, &num_colors, &(view->Width),
       &(view->Height), FI_MemAlloc, FI_WriteValue, FI_GIFerror);
       rows = 0;
       for(i=0; i<num_colors >>1; i++)
          {
           (XDC->colors)[i] = 2;
           ind1 = i<<1;
           ind2 = (i<<1)+1;
	   dum =  red[ind1];
       	   (XDC->beginR)[i] = (dum<<8)|dum;
	   dum =  green[ind1];
       	   (XDC->beginG)[i] = (dum<<8)|dum;
	   dum = blue[ind1];
       	   (XDC->beginB)[i] = (dum<<8)|dum;
	   dum =  red[ind2];
       	   (XDC->endR)[i] = (dum<<8)|dum;
	   dum =  green[ind2];
       	   (XDC->endG)[i] = (dum<<8)|dum;
	   dum =  blue[ind2];
       	   (XDC->endB)[i] = (dum<<8)|dum;
       	   rows++;
          }
       XDC->rowNum = rows;
       GR_InitColorImport( num_colors, num_colors, XDC, HWI);
       strcpy(view->DimX, "pixel");
       strcpy(view->DimY, "pixel");
       strcpy(view->DimZ, "color");
       view->FactX = 1.0;
       view->FactY = 1.0;
       view->OffX = 0.0;
       view->OffY = 0.0;
       view->links = 0;
       return(OK);
      }
    strcpy(XDim->error, XDERR_NO_VIEW);
    return(ERROR);
   }

/*
 * Remove Links after a viewport was deleted
 */

void FI_RemoveWasteLinks(int ViewNum, XDimInfo *XDim)
   {
    ViewportInfo *view;
    int i, j;

    for(i=0; i<MAX_VIEWPORT_NUMBER; i++)
       {
        view = &((XDim->Views)[i]);
        if(view->pict != NULL)
          {
           for(j=0; j<view->links; j++)
              {
               if((view->Link1)[j] == ViewNum || (view->Link2)[j] == ViewNum)
                  FI_DeleteLink(j, view, XDim);
              }
          }
       }
   }


/*
 * Delete a link from viewport
 */
   
void FI_DeleteLink(int ViewNum, ViewportInfo *view, XDimInfo *XDim)
   {
    int Num;

    for(Num= ViewNum+1; Num<view->links; Num++)
       {
	strncpy((view->LinkItself)[Num-1],
		(view->LinkItself)[Num], MAX_LINK_LENGTH);
	strncpy((view->LinkName)[Num-1],
		(view->LinkName)[Num], MAX_LINK_LENGTH);
	(view->Link1)[Num-1] = (view->Link1)[Num];
	(view->Link2)[Num-1] = (view->Link2)[Num];
       }
    if(view->links > 0)
      (view->links)--;
    view->actLink = view->links -1;
   }

/*
 * Return the port number to the link desciption
 */
 
int FI_GetLink(char *name, ViewportInfo *view, XDimInfo *XDim)
  {
   int i, port;
   
   port = -1;
   for(i=0; i<view->links; i++)
      {
       if(!strncmp((view->LinkName)[i], name, MAX_LINK_LENGTH))
         {
          port = (view->Link2)[i];
          break;
         }       
       if(!strncmp((view->LinkItself)[i], name, MAX_LINK_LENGTH))
         {
          port = (view->Link1)[i];
          break;
         }       
      }
   if(port <0)
     sprintf(XDim->error, XDERR_FIND_LINK_S, name);
   return(port);
  }

/*
 * Add link to viewport
 */

void FI_AddLink(int from, int to, char *fromName, char *toName, XDimInfo *XDim)
  {
   ViewportInfo *view1, *view2;
   int Num;
   
   view1 = &((XDim->Views)[from]);
   view2 = &((XDim->Views)[to]);
   Num = view1->links;
   strncpy((view1->LinkItself)[Num], fromName, MAX_LINK_LENGTH);
   strncpy((view1->LinkName)[Num], toName, MAX_LINK_LENGTH);
   (view1->Link1)[Num] = from;
   (view1->Link2)[Num] = to;
   (view1->links)++;
   Num = view2->links;
   strncpy((view2->LinkItself)[Num], toName, MAX_LINK_LENGTH);
   strncpy((view2->LinkName)[Num], fromName, MAX_LINK_LENGTH);
   (view2->Link2)[Num] = from;
   (view2->Link1)[Num] = to;
   (view2->links)++;
  }

/*
 * Export VRML data
 */

int FI_VRMLExport(char *FName, XDimInfo *XDim)
   {
    ViewportInfo *view;
    XDColorInfo *XDC;
    XDHardwareInfo *HWI;
    FILE *out;

    if(XDim->port >=0)
      {
       view = &((XDim->Views)[XDim->port]);
       XDC = &(XDim->colors);
       HWI = &(XDim->hwInfo);
       /*
        * init output
        */
       out = VR_InitVRMLout(FName);
       if(out)
         {
          /*
           * write palette
           */
          VR_WritePalette(out, 0.2, XDC->wantedColors, (HWI->red),
		(HWI->green), (HWI->blue));
	  /*
	   * write data set
	   */
	  switch(view->dataType)
	     {
	      case MY_DOUBLE:
	      VR_WriteDouble(out, view->Width, view->Height, view->OffX,
		view->FactX, view->OffX, view->FactY, view->Max, view->Min,
		view->actData, view->deco);
	      break;
	      default:
	      break;
	     }
          /*
           * close output
           */
          VR_CloseVRMLout(out);
          return(OK);
         }
       else
         {
	  sprintf(XDim->error, XDERR_OPEN_WRITE_S, FName);
	  return(ERROR);
         }
      }
    strcpy(XDim->error, XDERR_NO_VIEW);
    return(ERROR);
   }
