/****************************************************************/
/*								*/
/* obj3d.c:	Manipulation of 3D objects			*/
/*								*/
/*		Design: Walter Benzing 1995			*/
/*								*/
/*		NO RIGHTS RESERVED				*/
/****************************************************************/

/* standard includes */

#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <ctype.h>
#include <string.h>

/* my includes */

#include "xdim.h"
#include "obj3d.h"

/*
 * find a number,
 * dir gives direction
 */

char *O3_FindNumber(double *val, char *buf, char *start, int up)
   {
    char ch, *sp;

    if(buf == start)
      {
       if(up)
	 {
	  /* first call up => ignore token */
	  while((ch = *buf) != '\0' && ch != '+' && ch != '-' && !isdigit(ch))
		buf++;
	 }
       else
         {
          /* first call with down => add length */
          buf = start + strlen(start)-1;
         }
      }
    if(*buf == '\0' || buf < start)
      return(NULL);
    if(*buf == '!')
      {
       if(up)
         buf++;
       else
         buf--;
      }
    /* remove waste */
    while((ch = *buf) != '\0' && buf >= start)
       {
	if(ch != ' ' && ch != '\t' && ch != ',' && ch != ';'
		&& ch != '(' && ch != ')' && ch != '\n')
	  break;
	if(up)
	  buf++;
	else
          buf--;
       }
    if(!up)
      {
       while(buf >= start)
	 {
	  ch = *buf;
	  if(!isdigit(ch) && ch != '.' && ch != '+' && ch != '-' && ch != 'e' )
            break;
	  buf --;
	 }
       buf++;
      }
    if(*buf != '!' && *buf != '\0')
      {
       *val = strtod(buf, &sp);
       if(sp==buf)
	 buf= NULL;
       else if(up)
	 buf= sp;
       else
         buf--;
      }
    if(buf < start)
      buf = NULL;
    return(buf);
   }

/*
 * Insert a Polygon
 * ! means to start a new outline, p.ex. a plane
 * with a hole in it
 */

int O3_InsertPoly(char *start, int Length, double **data, int *dnum, int *free,
   int *num_col, unsigned char *red, unsigned char *green, unsigned char *blue,
   XDColorInfo *XDC, XDimInfo *XDim)
   {
    int xyz, anum, num, ind, status, p1[2*MAX_POLY], p2[2*MAX_POLY];
    double val, xd[2*MAX_POLY], yd[2*MAX_POLY], zd[2*MAX_POLY], col[2*MAX_POLY];
    double valz, val2, val3;
    char *c1, *c2, *c3, *buf;
    unsigned char rg, bg, gg;

    /*
     * extract color information first
     */
    buf = start;
    status = OK;
    if(XDC->opt3d & XD_3D_ONE_IND)
      {
       /*
        * one color index per polygon
        */
       c1 = O3_FindNumber(&valz, buf, start, 0);
       if(c1)
          *c1 = '\0';
       else
	 {
	  start[64] = '\0';
	  sprintf(XDim->error, XDERR_SYNTAX_S, start);
          return(ERROR);
         }
      }
    else if(XDC->opt3d & XD_3D_RGB)
      {
       /*
        * in rgb mode last 3 = color
        */
       c1 = O3_FindNumber(&valz, buf, start, 0);
       c2 = O3_FindNumber(&val2, c1, buf, 0);
       c3 = O3_FindNumber(&val3,c2, buf, 0);
       if(c1 && c2 && c3)
         {
          /*
           * don 't treat rgb as x,y,z
           */
          *c3 = '\0';
	  if(val3 < 1.0)
            rg = (unsigned char)(255.0*val3);
          else
            rg = (unsigned char) val3;
	  if(val2 < 1.0)
            gg = (unsigned char)(255.0*val2);
          else
            gg = (unsigned char) val2;
	  if(valz < 1.0)
            bg = (unsigned char)(255.0*valz);
          else
            bg = (unsigned char) valz;
          ind = GR_MatchColor(rg, gg, bg, &num, *num_col, red, green, blue);
	  if(*num_col < MAX_COL_NUMBER && num > COLOR_MATCH)
	    {
	     ind = *num_col;
	     red[ind] = rg;
	     green[ind] = gg;
	     blue[ind] = bg;
             (*num_col)++;
            }
         }
       else
         {
	  start[64] = '\0';
	  sprintf(XDim->error, XDERR_SYNTAX_S, start);
          return(ERROR);
         }
      }
    num = 0;
    anum = 0;
    xyz = 0;
    buf = start;
    while((buf = O3_FindNumber(&val, buf, start, True)) != NULL)
       {
        if(*buf == '!' || *buf == '\0')
          {
           if(anum > 2 && xyz ==0)
             {
              p1[num-anum] = num-1;
              p2[num-1] = num-anum;
              anum = 0;
             }
           else
             {
	      num = 0;
              status = ERROR;
	      start[64] = '\0';
	      sprintf(XDim->error, XDERR_SYNTAX_S, start);
              break;
             }
          }
        else
          {
           switch(xyz)
             {
              case 0:
                xd[num] = val;
                break;
              case 1:
                yd[num] = val;
                break;
              case 2:
                zd[num] = val;
             }
           xyz++;
          }
        if(xyz == 3)
          {
           if(anum)
             {
              p2[num-1] = num;
              p1[num] = num-1;
             }
           if(XDC->opt3d & XD_3D_RGB)
             col[num] = (double)ind;
           else if(XDC->opt3d & XD_3D_ZCOL)
             col[num] = zd[num];
           else if(XDC->opt3d & XD_3D_PER_VALUE)
             {
              buf = O3_FindNumber(&valz, buf, start, True);
              if(buf == NULL || *buf == '\0' || *buf == '!')
                {
		 num = 0;
                 status = ERROR;
		 start[64] = '\0';
		 sprintf(XDim->error, XDERR_SYNTAX_S, start);
		 break;
                }
              else
                col[num] = valz;
             }
           else
             col[num] = valz;
           xyz = 0;
           num++;
           anum++;
          }
	if(num == MAX_POLY)
	  break;
       }
    if(num)
      {
       /*
        * Transform polygon into a set of triangles
	*/
       if(O3_Poly2Triangle(xd, yd, zd, col, num, p1, p2,
		data, free, dnum))
	 {
	  status = ERROR;
          sprintf(XDim->error, XDERR_NO_MEM);
	 }

      }
    return(status);
   }


/*
 * Insert an Ellipsoid
 * if the number of points per full circle is missing
 * 36 is taken as default
 */

int O3_InsertEll(char *start, int Length, double **data, int *dnum, int *freeP,
   int *num_col, unsigned char *red, unsigned char *green, unsigned char *blue,
   FILE *in, XDColorInfo *XDC, XDimInfo *XDim)
   {
    int ind, i, j, status, perC, size, num;
    double val, x0, y0, z0, *col, r1, r2, r3;
    double valz, val2, val3, *dp, repl;
    char *c1, *c2, *c3, *buf;
    unsigned char rg, bg, gg;

    /*
     * extract color information first
     */
    buf = start;
    status = OK;
    if(XDC->opt3d & XD_3D_ONE_IND)
      {
       /*
        * one color index per polygon
        */
       c1 = O3_FindNumber(&valz, buf, start, 0);
       if(c1)
          *c1 = '\0';
       else
	 {
	  start[64] = '\0';
	  sprintf(XDim->error, XDERR_SYNTAX_S, start);
          return(ERROR);
         }
      }
    else if(XDC->opt3d & XD_3D_RGB)
      {
       /*
        * in rgb mode last 3 = color
        */
       c1 = O3_FindNumber(&valz, buf, start, 0);
       c2 = O3_FindNumber(&val2, c1, buf, 0);
       c3 = O3_FindNumber(&val3,c2, buf, 0);
       if(c1 && c2 && c3)
         {
          /*
           * don 't treat rgb as x,y,z
           */
          *c3 = '\0';
	  if(val3 < 1.0)
            rg = (unsigned char)(255.0*val3);
          else
            rg = (unsigned char) val3;
	  if(val2 < 1.0)
            gg = (unsigned char)(255.0*val2);
          else
            gg = (unsigned char) val2;
	  if(valz < 1.0)
            bg = (unsigned char)(255.0*valz);
          else
            bg = (unsigned char) valz;
          ind = GR_MatchColor(rg, gg, bg, &num, *num_col, red, green, blue);
	  if(*num_col < MAX_COL_NUMBER && num > COLOR_MATCH)
	    {
	     ind = *num_col;
	     red[ind] = rg;
	     green[ind] = gg;
	     blue[ind] = bg;
             (*num_col)++;
            }
	  valz = (double)ind;
         }
       else
         {
	  start[64] = '\0';
	  sprintf(XDim->error, XDERR_SYNTAX_S, start);
          return(ERROR);
         }
      }
    buf = start;
    for(i=0; i<7;i++)
       {
        buf = O3_FindNumber(&val, buf, start, True);
        if(i<6)
          {
           if(buf)
             {
              switch(i)
                {
                 case 0:
                 r1 = val;
                 break;
                 case 1:
                 r2 = val;
                 break;
                 case 2:
                 r3 = val;
                 break;
                 case 3:
                 x0 = val;
                 break;
                 case 4:
                 y0 = val;
                 break;
                 case 5:
                 z0 = val;
                 break;
                }
             }
          else
             {
  	      start[64] = '\0';
	      sprintf(XDim->error, XDERR_SYNTAX_S, start);
	      status = ERROR;
	      break;
             }
          }
        else
	  {
	   if(buf)
             {
	      perC = (int) val;
	      if(perC & 1 )
	        {
	         start[64] = '\0';
	         sprintf(XDim->error, XDERR_PC_EVEN_S, start);
                 status = ERROR;
		}
	      else if(perC < 3)
                {
		 start[64] = '\0';
	         sprintf(XDim->error, XDERR_PC_VAL_S, start);
                 status = ERROR;
		}
	     }
           else
             perC = 36;
          }
       }
    col = NULL;
    if((XDC->opt3d & XD_3D_PER_VALUE) && status == OK)
      {
       size = ((perC*perC)>>1)*sizeof(double);
       col = malloc(size);
       if(col)
         {
          /* read in the color data */
          size = (perC>>1);
          repl = 0.0;
          dp = col;
          i = 0;
          while(fgets(buf, Length, in) != NULL)
              {
               if(*buf != '#' && *buf !='"')
                 {
                  if(i==0 || i == size)
                    {
                     if((buf=FI_GetNextNum(buf, &val)) == NULL)
                       val = repl;
                     else
                       repl = val;
                     *dp++ = val;
                     if(i==size)
                       break;
		    }
		  else
                    {
                     for(j=0; j< perC; j++)
                        {
                         if((buf=FI_GetNextNum(buf, &val)) == NULL)
                	   val = repl;
                         else
                           repl = val;
			 *dp++ = val;
			}
                     }
                  i++;
                 }
              }
          if(i<size)
	    {
	     strcpy(XDim->error, XDERR_EOF);
	     status = ERROR;
	    }
         }
       else
         {
          sprintf(XDim->error, XDERR_NO_MEM_D, size);
          status = ERROR;
         }
      }
    if(status==OK)
      {
       if(O3_Ellips2Triangle(x0,y0,z0,r1,r2,r3,perC,XDC->opt3d,valz,col,
		data, dnum, freeP))
	 {
	  status = ERROR;
          sprintf(XDim->error, XDERR_NO_MEM);
	 }
      }
    if(col)
      free(col);
    return(status);
   }

/*
 * Insert unstructured 3d data
 * a point will be connected if it's inside delx, dely or delz
 */

int O3_Insert3D(char *start, int Length, double **data, int *dnum, int *freeP,
   int *num_col, FILE *in, XDColorInfo *XDC, XDimInfo *XDim)
   {
    int status, num,free3D,xyz;
    size_t size;
    double *xd, *yd, *zd, *col, val;
    long int last , pos;
    char *buf;

    status = OK;
    /* first read in the points */
    free3D = 0;
    num = 0;
    xd = NULL;
    yd = NULL;
    zd = NULL;
    col = NULL;
    pos = ftell(in);
    do
      {
       last = pos;
       pos = ftell(in);
       xyz = 0;
       buf = start;
       if(free3D==0)
	 {
	  size = (num+MAX_MAPPED)*sizeof(double);
	  xd = realloc(xd,size);
	  yd = realloc(yd,size);
	  zd = realloc(zd,size);
	  col = realloc(col,size);
          if(xd && yd && zd && col)
	    free3D += MAX_MAPPED;
	  else
	    {
	     free(xd);
	     free(yd);
	     free(zd);
	     free(col);
	     num = 0;
             return(ERROR);
            }
         }
       while((buf=FI_GetNextNum(buf, &val)) != NULL)
	  {
	   if(status)
	     {
              status = OK;
	      break;
             }
	   switch(xyz)
	     {
	      case 0:
	      xd[num] = val;
	      break;
	      case 1:
	      yd[num] = val;
	      break;
	      case 2:
	      zd[num] = val;
	      break;
	      case 3:
	      col[num] = val;
	      break;
	      default:
	      status = ERROR;
              break;
             }
	   xyz++;
          }
       if(xyz==3 && status==OK)
         {
	  col[num] = zd[num];
	  num++;
          free3D--;
         }
       else if(xyz==4 && status==OK)
         {
	  num++;
	  free3D--;
	 }         
       else
	 {
	  fseek(in, last, SEEK_SET);
	  break;
         }
      }
    while(fgets(start, Length, in) != NULL);
    if(num > 2)
      {
       if(O3_3D2Triangle(xd, yd, zd, col, XDC->opt3d, num, data, freeP, dnum))
	 {
	  status = ERROR;
          sprintf(XDim->error, XDERR_NO_MEM);
	 }
      }
    if(xd)
      free(xd);
    if(yd)
      free(yd);
    if(zd)
      free(zd);
    if(col)
      free(col);
    return(status);
   }

/*
 * Insert Lines
 * instead of lines triangles are used
 * this will change !
 */

int O3_InsertLines(char *start, int Length, double **data, int *dnum, int *free,
   int *num_col, unsigned char *red, unsigned char *green, unsigned char *blue,
   XDColorInfo *XDC, XDimInfo *XDim)
   {
    int ind, status, num, xyz;
    double val, xd[MAX_POLY], yd[MAX_POLY], zd[MAX_POLY];
    double valz, val2, val3;
    char *c1, *c2, *c3, *buf;
    unsigned char rg, bg, gg;

    /*
     * extract color information first
     */
    buf = start;
    status = OK;
    if(XDC->opt3d & XD_3D_RGB)
      {
       /*
        * in rgb mode last 3 = color
        */
       c1 = O3_FindNumber(&valz, buf, start, 0);
       c2 = O3_FindNumber(&val2, c1, buf, 0);
       c3 = O3_FindNumber(&val3,c2, buf, 0);
       if(c1 && c2 && c3)
         {
          /*
           * don 't treat rgb as x,y,z
           */
          *c3 = '\0';
	  if(val3 < 1.0)
            rg = (unsigned char)(255.0*val3);
          else
            rg = (unsigned char) val3;
	  if(val2 < 1.0)
            gg = (unsigned char)(255.0*val2);
          else
            gg = (unsigned char) val2;
	  if(valz < 1.0)
            bg = (unsigned char)(255.0*valz);
          else
            bg = (unsigned char) valz;
          ind = GR_MatchColor(rg, gg, bg, &num, *num_col, red, green, blue);
	  if(*num_col < MAX_COL_NUMBER && num > COLOR_MATCH)
	    {
	     ind = *num_col;
	     red[ind] = rg;
	     green[ind] = gg;
	     blue[ind] = bg;
             (*num_col)++;
            }
	  valz = (double)ind;
         }
       else
         {
	  start[64] = '\0';
	  sprintf(XDim->error, XDERR_SYNTAX_S, start);
          return(ERROR);
         }
      }
    else
      {
       /*
        * one color index per polygon
        */
       c1 = O3_FindNumber(&valz, buf, start, 0);
       if(c1)
          *c1 = '\0';
       else
	 {
	  start[64] = '\0';
	  sprintf(XDim->error, XDERR_SYNTAX_S, start);
          return(ERROR);
         }
      }
    /* now get the numbers but no more than MAX_POLY */
    num = 0;
    xyz = 0;
    buf = start;
    while((buf = O3_FindNumber(&val, buf, start, True)) != NULL)
       {
        if(*buf != '!' && *buf != '\0')
	  {
           switch(xyz)
             {
              case 0:
                xd[num] = val;
                break;
              case 1:
                yd[num] = val;
                break;
              case 2:
                zd[num] = val;
             }
           xyz++;
          }
        if(xyz == 3)
          {
	   xyz = 0;
           num++;
	   if(num == MAX_POLY)
	     break;
	  }
       }
    if(xyz)
      {
       num = 0;
       status = ERROR;
       start[64] = '\0';
       sprintf(XDim->error, XDERR_SYNTAX_S, start);
      }
    if(num)
      {
       /*
        * Transform polygon into a set of triangles
	*/
       if(O3_Lines2Lin(xd, yd, zd, valz, num, data, free, dnum))
	 {
	  status = ERROR;
          sprintf(XDim->error, XDERR_NO_MEM);
	 }
      }
    return(status);
   }

/*
 * Insert a Triangle into data stream
 */

int O3_InsertTriangle(double x0, double y0, double z0, double c0, double x1,
	double y1, double z1, double c1, double x2, double y2, double z2,
	double c2, double **data, int *dnum, int *free)
   {
    int del, size;
    double *dum;

    del = 12*(*dnum);
    if(*free < 12)
      {
       *free = MAX_POLY;
       size = (del+MAX_POLY)*sizeof(double);
       dum = realloc(*data, size);
       if(dum == NULL)
	  return(ERROR);
       *data = dum;
      }
    dum = (*data)+del;
    *free -= 12;
    *dnum += 1;
    *dum++ = x0;
    *dum++ = y0;
    *dum++ = z0;
    *dum++ = x1;
    *dum++ = y1;
    *dum++ = z1;
    *dum++ = x2;
    *dum++ = y2;
    *dum++ = z2;
    /* color info */
    *dum++ = c0;
    *dum++ = c1;
    *dum++ = c2;
    return(OK);
   }
/*
 * create triangles from unformated 3d
 */

int O3_3D2Triangle(double *xd, double *yd, double *zd, double *col,
	int opt3d, int num, double **data, int *freeP, int *dnum)
    {
     int status, lin1, lin2, nl1, nl2, lin, n0, n1, n2, n3, *p, i, j;
     int l1, l2, len1, len2, k;
     double *min, *maj;
     double x0, x1;

     status = OK;
     p = malloc(num*sizeof(int));
     if(p==NULL)
       return(ERROR);
     for(i=0; i<num;i++)
        p[i] = i;
     if(opt3d & XD_SORTXZ)
       {
        min = xd;
        maj = zd;
       }
     else if(opt3d & XD_SORTYZ)
       {
        min = yd;
        maj = zd;
       }
     else
       {
        min = xd;
        maj = yd;
       }
     if(opt3d & XD_SORT)
       O3_SortMM(min,maj,num,p);
     /* count the lines -1 */
     lin = 0;
     x1 = min[p[0]];
     for(i=1; i< num; i++)
        {
         x0 = x1;
         x1 = min[p[i]];
         if(x1 < x0)
           lin++;
        }
     lin2 = 0;
     /* search first line */
     x1 = min[0];
     nl2 = 1;
     /* search next line */
     for(i=1; i<num; i++)
        {
         x0 = x1;
         x1 = min[p[i]];
         if(x1 > x0)
           nl2++;
         else
           break;
        }
     for(i=0; i< lin; i++)
        {
         lin1 = lin2;
         nl1 = nl2;
         lin2 += nl2;
	 x1 = min[lin2];
	 nl2 = 1;
         /* search next line */
         for(j=lin2+1; j<num; j++)
            {
             x0 = x1;
             x1 = min[p[j]];
             if(x1 > x0)
               nl2++;
             else
               break;
            }
         if(nl2 > nl1)
           {
            len1 = nl2;
            len2 = nl1-1;
            l1 = lin2;
            l2 = lin1;
           }
         else
           {
            len1 = nl1;
            len2 = nl2-1;
            l1 = lin1;
            l2 = lin2;
           }
         k = l2;
         n1 = p[l1];
         n3 = p[k];
         for(j=1; j< len1; j++)
            {
             n0 = n1;
             n1 = p[l1+j];
             n2 = n3;
             if(k < l2+len2)
               k++;
             n3 = p[k];
	     if(O3_InsertTriangle(xd[n0], yd[n0], zd[n0], col[n0], xd[n1],
		yd[n1], zd[n1], col[n1], xd[n2], yd[n2], zd[n2], col[n2],
		data, dnum, freeP))
               {
        	status = ERROR;
		break;
	       }
	     if(n2!=n3)
	       {
		if(O3_InsertTriangle(xd[n1], yd[n1], zd[n1], col[n1], xd[n2],
		   yd[n2], zd[n2], col[n2], xd[n3], yd[n3], zd[n3], col[n3],
		   data, dnum, freeP))
        	  {
        	   status = ERROR;
		   break;
		  }
	       }
            }
          if(status)
            break;
        }
     free(p);
     return(status);
    }

/*
 * Create lines from raw data
 * at the moment triangles are inserted
 * this will change
 */

int O3_Lines2Lin(double *xd, double *yd, double *zd, double valz, int num,
   double **data, int *free, int *dnum)
  {
   int status,i;
   double x0,y0,z0,x1,y1,z1;

   status = OK;
   x1 = *xd;
   y1 = *yd;
   z1 = *zd;
   for(i=1; i<num; i++)
      {
       x0 = x1;
       y0 = y1;
       z0 = z1;
       x1 = xd[i];
       y1 = yd[i];
       z1 = zd[i];
       if(O3_InsertTriangle(x0, y0, z0, valz, x1, y1, z1, valz,
		x1, y1, z1, valz, data, dnum, free))
         {
          status = ERROR;
	  break;
	 }
      }
   return(status);
  }


/*
 * create an ellipsoid from raw data
 */

int O3_Ellips2Triangle(double x0,double y0,double z0, double r1, double r2,
   double r3, int perC, int opt3d, double valz, double *col, double **data,
   int *dnum, int *free)
   {
    int status,i,j,end;
    double x1,y1,z1,x2,y2,z2,x3,y3,z3,dz,x4,y4,z4,dPhi,Phi1, Phi2;
    double ca,cb,sa,sb,dr1,dr2,c1,c2,c3,c4,c5,c6,*cp1,*cp2,*cl1,*cl2;
    double dr3,dr4,*cl4;
    
    status = OK;
    /* start with first and last point*/
    dPhi = 2*PI/perC;
    x1 = x0;
    y1 = y0;
    z1 = z0-r3;
    z2 = z0+r3;
    dz = r3*cos(dPhi);
    z3 = z0-dz;
    z4 = z0+dz;
    sa = sin(dPhi);
    dr1 = r1*sa;
    dr2 = r2*sa;
    Phi2 = 0.0;
    if(opt3d & XD_3D_ONE_IND || (opt3d&XD_3D_RGB))
      {
       c1 = valz;
       c2 = valz;
       c3 = valz;
       c4 = valz;
       c5 = valz;
       c6 = valz;
      }
    else if(opt3d&XD_3D_ZCOL)
      {
       c1 = z1;
       c2 = z2;
       c3 = z3;
       c4 = z3;
       c5 = z4;
       c6 = z4;
      }
    else if(opt3d&XD_3D_PER_VALUE)
      {
       cl1 = col;
       cl2 = cl1+1;
       cl4 = col+1+((perC>>1)-1)*perC;
       c1 = *cl1;
       c2 = *cl4;
       cl4 -= perC;
       cp1 = cl4;
       cp2 = cl2;
       c4 = *cp2++;
       c6 = *cp1++;
      }
    x4 = dr1+x0;
    y4 = y0;
    /* insert first and last row */
    for(i=1; i<=perC;i++)
       {
        Phi2 += dPhi;
        cb= cos(Phi2);
        sb= sin(Phi2);
        x3 = x4;
        y3 = y4;
        x4 = dr1*cb+x0;
        y4 = dr2*sb+y0;
	if(opt3d&XD_3D_PER_VALUE)
	  {
	   c3 = c4;
	   c5 = c6;
	   if(i<perC)
	     {
              c4 = *cp2++;
	      c6 = *cp1++;
	     }
	   else
	     {
	      c4 = *cl4;
	      c6 = *cl2;
	     }
          }
	if(O3_InsertTriangle(x1, y1, z1, c1, x3, y3, z3, c3,
		x4, y4, z3, c4, data, dnum, free))
	  {
	   status = ERROR;
	   break;
	  }
	if(O3_InsertTriangle(x1, y1, z2, c2, x3, y3, z4, c5,
		x4, y4, z4, c6, data, dnum, free))
	  {
	   status = ERROR;
	   break;
	  }
       }
    /* insert middle rows */
    Phi1 = dPhi;
    end = perC>>1;
    sa = sin(Phi1);
    ca = cos(Phi1);
    dr3 = r1*sa;
    dr4 = r2*sa;
    z3 = z0-r3*ca;
    z4 = z3;
    for(i=2;i<end;i++)
       {
        dr1 = dr3;
        dr2 = dr4;
        z1 = z3;
        z2 = z4;
        Phi1 += dPhi;
        sa = sin(Phi1);
        ca = cos(Phi1);
        dr3 = r1*sa;
        dr4 = r2*sa;
        z3 = z0-r3*ca;
        z4 = z3;
        Phi2 = 0.0;
	x2 = dr1+x0;
	y2 = y0;
	x4 = dr3+x0;
	y4 = y0;
	if(opt3d&XD_3D_ZCOL)
	  {
	   c1 = z1;
	   c2 = z2;
	   c3 = z3;
	   c4 = z4;
	  }
	else if(opt3d&XD_3D_PER_VALUE)
	  {
	   cl1 = cl2;
	   cl2 = cp2;
	   cp1 = cl1;
	   c2 = *cp1++;
	   c4 = *cp2++;
	  }
	for(j=1; j<=perC;j++)
	   {
            Phi2 += dPhi;
            cb= cos(Phi2);
            sb= sin(Phi2);
            x1 = x2;
            y1 = y2;
            x2 = dr1*cb+x0;
            y2 = dr2*sb+y0;
            x3 = x4;
            y3 = y4;
            x4 = dr3*cb+x0;
            y4 = dr4*sb+y0;
	    if(opt3d&XD_3D_PER_VALUE)
	      {
	       c1 = c2;
	       c3 = c4;
	       if(j<perC)
		 {
        	  c2 = *cp1++;
		  c4 = *cp2++;
		 }
	       else
		 {
		  c2 = *cl1;
		  c4 = *cl2;
		 }
              }
	    if(O3_InsertTriangle(x1, y1, z1, c1, x3, y3, z3, c3,
		    x4, y4, z4, c4, data, dnum, free))
	      {
	       status = ERROR;
	       break;
	      }
	    if(O3_InsertTriangle(x1, y1, z1, c1, x2, y2, z2, c2,
		    x4, y4, z4, c4, data, dnum, free))
	      {
	       status = ERROR;
	       break;
	      }
	   }
       }
    return(status);
   }


/*
 * Arrays needed for sorting in Major and
 * Minor direction, MinDim, MajDim => smallest feature
 * PL1, PL2 link to next points in polygon
 */

double *Major3D, *Minor3D, MinDim, MajDim, Min0, Maj0;
int *PL1, *PL2;

/*
 * sort the indices in the polygon
 * WARNING: if Max-Min gets to large
 * this will fail;
 */

int O3_sortIndex(const void *index1,const void *index2)
  {
   double del,x1,x2,y1,y2,delx1,dely1,delx2,dely2, delx3, dely3, delx4, dely4;
   const int *ind1, *ind2;
   int ret, i1, i2, l1, l2, l3, l4;
   
   ind1 = index1;
   ind2 = index2;
   i1 = *ind1;
   i2 = *ind2;
   l1 = *(PL1+i1);
   l2 = *(PL2+i1);
   l3 = *(PL1+i2);
   l4 = *(PL2+i2);
   if(l1 < 0 || l2 < 0)
     ret = -1;
   else if(l3 < 0 || l4 < 0)
     ret = 1;
   else
     {
      y1 =  *(Major3D+i1);
      y2 =  *(Major3D+i2);
      del = y2-y1;
      if(del > MajDim)
	ret = 1;
      else if(del < -MajDim)
	ret = -1;
      else
	{
	 x1 =  *(Minor3D+i1);
	 x2 =  *(Minor3D+i2);
	 del = x2 - x1;
	 if(del > MinDim)
           ret = 1;
	 else if(del < -MinDim)
           ret = -1;
	 else
	   {
	    /* now check the next points */
	    y1 -= MajDim;
	    y2 -= MajDim;
	    dely1 = *(Major3D+l1)-y1;
	    dely2 = *(Major3D+l2)-y1;
	    dely3 = *(Major3D+l1)-y2;
	    dely4 = *(Major3D+l2)-y2;
	    if(dely1 < 0 || dely2 <0 || dely3 < 0 || dely4<0)
	      {
	      /*don't know and care */
	      if(y2>y1)
	        ret = 1;
	      else
	        ret = -1;
	      }
	    else
	      {
	       delx1 = *(Major3D+l1)-x1;
	       delx2 = *(Major3D+l2)-x1;
	       delx3 = *(Major3D+l1)-x2;
	       delx4 = *(Major3D+l2)-x2;
	       if(dely1*delx2 > dely2*delx1)
	         {
		  if(dely3*delx4 > dely4*delx3)
		    {
		     if(dely1*delx3 > dely3*delx1
			&& dely4*delx2 > dely2*delx4)
			ret = 1;
		     else
                        ret = -1;
		    }
		  else
	            {
		     if(dely1*delx4 > dely4*delx1
			&& dely3*delx2 > dely2*delx3)
			ret = 1;
		     else
                        ret = -1;
		    }
	         }
	       else
	         {
		  if(dely3*delx4 > dely4*delx3)
	            {
		     if(dely2*delx3 > dely3*delx2
			&& dely4*delx1 > dely1*delx4)
			ret = 1;
		     else
                        ret = -1;
		    }
		  else
	            {
		     if(dely2*delx4 > dely4*delx2
			&& dely3*delx1 > dely1*delx3)
			ret = 1;
		     else
                        ret = -1;
		    }
	         }
	      } 
	   }
	}
     }
   return(ret);
  }

/*
 * if more than one point is inside
 * a triangle, these points must be sorted in minor direction
 */

int O3_sortMinor(const void *index1,const void *index2)
  {
   double val1, val2;
   const int *ind1, *ind2;
   int ret, i1, i2;
   
   ind1 = index1;
   ind2 = index2;
   i1 = *ind1;
   i2 = *ind2;
   val1 = *(Minor3D+i1);
   val2 = *(Minor3D+i2);
   if(val1 < val2)
     ret = -1;
   else
     ret = 1;
   return(ret);
  }

/*
 * if more than one point is inside
 * a triangle, these points must be sorted in minor direction
 */

int O3_sortMinMaj(const void *index1,const void *index2)
  {
   double val1, val2;
   const int *ind1, *ind2;
   int ret, i1, i2;
   
   ind1 = index1;
   ind2 = index2;
   i1 = *ind1;
   i2 = *ind2;
   val1 = *(Major3D+i1);
   val2 = *(Major3D+i2);
   if(val1 < val2)
     ret = -1;
   else if(val1 > val2)
     ret = 1;
   else
     {
      val1 = *(Minor3D+i1);
      val2 = *(Minor3D+i2);
      if(val1 < val2)
	ret = -1;
      else
	ret = 1;
     }
   return(ret);
  }
/*
 * sort by angle
 */

int O3_sortDelxDely(const void *index1,const void *index2)
  {
   double x1, y1,x2,y2;
   const int *ind1, *ind2;
   int ret, i1, i2;
   
   ind1 = index1;
   ind2 = index2;
   i1 = *ind1;
   i2 = *ind2;
   x1 = *(Minor3D+i1);
   y1 = *(Major3D+i1);
   y2 = *(Major3D+i2);
   x2 = *(Minor3D+i2);
   if((y1-Maj0)*(x2-Min0) > (y2-Maj0)*(x1-Min0))
     ret = 1;
   else
     ret = -1;
   return(ret);
  }

/*
 * removes points from the list and inserts triangles into data
 * returns: error status
 * ma, mi:  projected plane data (typ. xd=Minor and yd=Major)
 * xd, yd, zd: plane data
 * p1, p2:  links to next points in a polygon
 * index:   data index: unsorted->sorted
 * rindex:  data index: sorted->unsorted
 */

int O3_NextTriangle(double *ma, double *mi, double *xd, double *yd, double *zd,
   double *col, double dimx, double dimy, int *p1, int *p2, int *index,
   int *rindex, int *nump, double **data, int *dnum, int *free)
   {
    double x, y, x0, y0, x1, y1, x2, y2, delx1, dely1, delx2, dely2;
    double delx3, dely3, delx4, dely4, delx, f1, f2, f3;
    double xs, ys, zs, cs, xs1, xs2, ys1, ys2, zs1, zs2, cs1, cs2, minM;
    double x5,y5,x6,y6;
    int num, n, i0, i, i1, i2, l1, l2, inside[2*MAX_POLY], count, status;
    int i3, i4, stat, ins, i5, i6;

    status = OK;
    num = (*nump) -1;
    /* get triangle data */
    i0 = rindex[num];
    x0 = *(mi+i0);
    y0 = *(ma+i0);
    i1 = *(p1+i0);
    i2 = *(p2+i0);
    if(i1 <0 || i2 <0)
      {
       /*This point was removed before */
       if(num <= 2)
          *nump = 0;
       else
	 *nump = num;
       return(OK);
      }
    l1 = index[i1];
    l2 = index[i2];
    if(l2 > l1)
      {
       i = l1;
       l1 = l2;
       l2 = i;
       i = i1;
       i1 = i2;
       i2 = i;
      }
    x1 = *(mi+i1);
    y1 = *(ma+i1);
    x2 = *(mi+i2);
    y2 = *(ma+i2);
    i3= 0;
    i4= 0;
    delx1 = x1-x0;
    if(fabs(delx1)<dimx)
      i3++;
    dely1 = y1-y0;
    if(dely1<dimy)
      i4++;
    delx2 = x2-x0;
    if(fabs(delx2)<dimx)
      i3++;
    dely2 = y2-y0;
    if(dely2<dimy)
      i4++;
    delx3 = x2-x1;
    if(fabs(delx3)<dimx)
      i3++;
    dely3 = y2-y1;
    if(dely3<dimy)
      i4++;
    if(i3>1 || i4>1)
      {
       /* Throw it away */
       O3_CLOSE_LINE(p1,p2,i0,i1,i2);
       O3_CLOSE_LINE(p1,p2,i0,i2,i1);
       if(num <= 2)
          *nump = 0;
       else
	  *nump = num;
       return(OK);
      }
    if(dely1 != 0.0)
      f1 = delx1/dely1;
    else
      f1 = 0.0;
    if(dely2 != 0.0)
      f2 = delx2/dely2;
    else
      f2 = 0.0;
    if(dely3 != 0.0)
      f3 = delx3/dely3;
    else
      f3 = 0.0;
    count = 0;
    minM = y2;
    /* now start testing */
    for(i=l2+1; i<num; i++)
       {
        if(i!=l1)
          {
	   n = rindex[i];
	   x = *(mi+n);
	   y = *(ma+n);
	   if(y > y0-dimy && y < y2-dimy && *(p1+n) >= 0 && *(p2+n) >= 0)
	     {
	      if(y<y1)
		{
		 xs1 = f1*(y-y1)+x1;
		 xs2 = f2*(y-y2)+x2;
		}
	      else
		{
		 xs1 = f3*(y-y1)+x1;
		 xs2 = f2*(y-y2)+x2;
		}
	      if(xs2< xs1)
		{
		 delx = xs1;
		 xs1 = xs2;
                 xs2 = delx;
                 stat = -1;
                }
              else
                stat = 1;
	      if(x >= xs1-dimx && x <=xs2+dimx)
		{
		 if(x >= xs1+dimx && x <=xs2-dimx)
	           {
	            inside[count] = n;
		    count++;
		    if(y<minM)
                      minM = y;
		   }
	         else
	           {
	            /*
	             *UUps near the edges =>
	             *test if overlap
	             */
		    i3 = *(p1+n);
		    i4 = *(p2+n);
		    i5 = -1;
		    i6 = -1;
		    x5 = *(mi+i3);
		    y5 = *(ma+i3);
		    x6 = *(mi+i4);
		    y6 = *(ma+i4);
		    ins = True;
		    if(x < xs1+dimx)
		      {
		       if(stat >0)
	                 {
	                  if(y<y1+dimy)
	                    {
	                     /* left of i1 */
			     xs = f1*(y5-y1)+x1;
			     if(x5 < xs+dimx)
                               i5 = i3;
			     xs = f1*(y6-y1)+x1;
			     if(x6 < xs+dimx)
			       {
			        if(i5 <0)
			          i5 = i4;
			        else
				  ins = False;
			       }
			    }
	                  if(y>y1-dimy)
	                    {
	                     /* left of i3 */
			     xs = f3*(y5-y1)+x1;
			     if(x5 < xs+dimx)
			       i6 = i3;
			     xs = f3*(y6-y1)+x1;
			     if(x6 < xs+dimx)
			       {
			        if(i6 <0)
			          i6 = i4;
			        else
 				  ins = False;
                               }
			    }
			  if(ins && i5>=0 && i6>=0 && i5 != i6 &&
				x < x1+dimx && x > x1-dimx)
			    {
			     /* sits on p1 */
			     x5 = *(mi+i5);
			     y5 = *(ma+i5);
			     x6 = *(mi+i6);
			     y6 = *(ma+i6);
			     if(y5>y6)
			       {
			        if(y5 > y1+dimy)
			          {
			           xs = (x5-x1)/(y5-y1)*(y6-y1)+x1;
			           if(x6 < xs+dimx)
			             ins = False;
			          }
			       }
			     else
			       {
			        if(y6 > y1+dimy)
			          {
			           xs = (x6-x1)/(y6-y1)*(y5-y1)+x1;
			           if(x5 < xs+dimx)
			             ins = False;
			          }
			       }
			    }
	                 }
	               else
	                 {
	                  /* left of i2 */
			  xs = f2*(y5-y2)+x2;
			  if(x5 < xs+dimx)
                            {
			     xs = f2*(y6-y2)+x2;
			     if(x6 < xs+dimx)
				ins = False;
                            }
			 }
	              }
	            if(x > xs2-dimx && ins)
		      {
		       if(stat < 0)
	                 {
	                  if(y<y1+dimy)
	                    {
	                     /* right of i1 */
			     xs = f1*(y5-y1)+x1;
			     if(x5 > xs-dimx)
				i5=i3;
			     xs = f1*(y6-y1)+x1;
			     if(xs > xs-dimx)
                               {
			        if(i5 <0)
			          i5 = i4;
			        else
				  ins = False;
                               }
			    }
	                  if(y>y1-dimy)
	                    {
	                     /* right of i3 */
			     xs = f3*(y5-y1)+x1;
			     if(x5 > xs-dimx)
				i6=i3;
			     xs = f3*(y6-y1)+x1;
			     if(x6 > xs-dimx)
                               {
			        if(i6 <0)
			          i6 = i4;
			        else
 				  ins = False;
                               }
	                    }
			  if(ins && i5>=0 && i6>=0 && i5 != i6 &&
				x < x1+dimx && x > x1-dimx)
			    {
			     /* sits on p1 */
			     x5 = *(mi+i5);
			     y5 = *(ma+i5);
			     x6 = *(mi+i6);
			     y6 = *(ma+i6);
			     if(y5>y6)
			       {
			        if(y5 > y1+dimy)
			          {
			           xs = (x5-x1)/(y5-y1)*(y6-y1)+x1;
			           if(x6 > xs-dimx)
			             ins = False;
			          }
			       }
			     else
			       {
			        if(y6 > y1+dimy)
			          {
			           xs = (x6-x1)/(y6-y1)*(y5-y1)+x1;
			           if(x5 > xs-dimx)
			             ins = False;
			          }
			       }
			    }
	                 }
	               else
	                 {
	                  /* right of i2 */
			  xs = f2*(y5-y2)+x2;
			  if(x5 > xs-dimx)
			    {
			     xs = f2*(y6-y2)+x2;
			     if(x6 > xs-dimx)
 				ins = False;
                            }
			 }
	              }
		    if(ins)
		      {
	               inside[count] = n;
	               count++;
		       if(y<minM)
			 minM = y;
		      }
	           }
	        }
	     }
	  }
       }
    if(count)
      {
       /* points inside */
       if(count == 1)
         {
          /* only one inside */
          n = *inside;
	  if(O3_InsertTriangle(*(xd+i0), *(yd+i0), *(zd+i0), *(col+i0),
	   *(xd+n), *(yd+n), *(zd+n), *(col+n), *(xd+i1), *(yd+i1),
	   *(zd+i1), *(col+i1), data, dnum, free))	status = ERROR;
	  else
	    {
	     if(O3_InsertTriangle(*(xd+i0), *(yd+i0), *(zd+i0), *(col+i0),
	      *(xd+n), *(yd+n), *(zd+n), *(col+n), *(xd+i2), *(yd+i2),
	      *(zd+i2), *(col+i2), data, dnum, free))	status = ERROR;
	     /* remove point and insert other*/
	     x = *(mi+n);
	     y = *(ma+n);
	     *(xd+i0) = *(xd+n);
	     *(yd+i0) = *(yd+n);
	     *(zd+i0) = *(zd+n);
	     *(col+i0) = *(col+n);
	     i3 = *(p1+n);
	     i4 = *(p2+n);
	     delx3 = *(mi+i3)-x;
	     dely3 = *(ma+i3)-y;
	     delx4 = *(mi+i4)-x;
	     dely4 = *(ma+i4)-y;
	     /* now break up into 2 */
	     if(dely1*delx2 > dely2*delx1)
	       {
	        /* p1->p2 */
		*(p1+i0) = i1;
		if(*(p1+i2) == i0)
		  *(p1+i2) = n;
		else
		  *(p2+i2) = n;
		if(dely3*delx4 > dely4*delx3)
		  {
		   /* p3->p4 */
		   *(p2+i0) = i3;
		   *(p1+n) = i2;
		   if(*(p1+i3) == n)
		     *(p1+i3) = i0;
		   else
		     *(p2+i3) = i0;
		  }
		else
		  {
		   /* p4->p3 */
		   *(p2+i0) = i4;
		   *(p2+n) = i2;
		   if(*(p1+i4) == n)
		     *(p1+i4) = i0;
		   else
		     *(p2+i4) = i0;
		  }
	       }
	     else
	       {
	        /* p2->p1 */
		*(p1+i0) = i2;
		if(*(p1+i1) == i0)
		  *(p1+i1) = n;
		else
		  *(p2+i1) = n;
		if(dely3*delx4 > dely4*delx3)
		  {
		   /* p3->p4 */
		   *(p2+i0) = i3;
		   *(p1+n) = i1;
		   if(*(p1+i3) == n)
		     *(p1+i3) = i0;
		   else
		     *(p2+i3) = i0;
		  }
		else
		  {
		   /* p4->p3 */
		   *(p2+i0) = i4;
		   *(p2+n) = i1;
		   if(*(p1+i4) == n)
		     *(p1+i4) = i0;
		   else
		     *(p2+i4) = i0;
		  }
	        
	       }
	     if(num <= 3)
        	*nump = 0;	     
            }
         }
       else
         {
	  /*
           * count > 1
	   */
	  /* Now sort the extra points */
	  qsort(inside, count, sizeof(int), O3_sortMinor);
	  if(dely2 != 0.0)
	    f2 = (minM- y0)/dely2;
	  else
	    f2 = 1.0;
	  xs2 = f2*(*(xd+i2)- *(xd+i0))+ *(xd+i0);
	  ys2 = f2*(*(yd+i2)- *(yd+i0))+ *(yd+i0);
	  zs2 = f2*(*(zd+i2)- *(zd+i0))+ *(zd+i0);
	  cs2 = f2*(*(col+i2)- *(col+i0))+ *(col+i0);
	  xs = f2*(*(mi+i2)- *(mi+i0))+ *(mi+i0);
	  if(minM > y1+dimy)
	    {
	     xs1 = *(xd+i1);
	     ys1 = *(yd+i1);
	     zs1 = *(zd+i1);
	     cs1 = *(col+i1);
	     x = *(mi+i1);
	     y = minM+dimy;
	     for(i=count-1; i>=0; i--)
	        {
	         i4 = inside[i];
	         if(*(ma+i4) < y)
	           break;
	        }
	     if(O3_InsertTriangle(*(xd+i0), *(yd+i0), *(zd+i0), *(col+i0),
		xs1, ys1, zs1, cs1, xs2, ys2, zs2, cs2,
		data, dnum, free))	status = ERROR;
	     else if(O3_InsertTriangle(xs1, ys1, zs1, cs1,
		*(xd+i4), *(yd+i4), *(zd+i4), *(col+i4), xs2, ys2, zs2, cs2,
			data, dnum, free))
		status = ERROR;
	    }
	  else
	    {
	     if(dely1 != 0.0)
	       f1 = (minM- y0)/dely1;
	     else
	       f1 = 1.0;
	     xs1 = f1*(*(xd+i1)- *(xd+i0))+ *(xd+i0);
	     ys1 = f1*(*(yd+i1)- *(yd+i0))+ *(yd+i0);
	     zs1 = f1*(*(zd+i1)- *(zd+i0))+ *(zd+i0);
	     cs1 = f1*(*(col+i1)- *(col+i0))+ *(col+i0);
	     x = f1*(*(mi+i1)- *(mi+i0))+ *(mi+i0);
	     if(O3_InsertTriangle(*(xd+i0), *(yd+i0), *(zd+i0), *(col+i0),
		xs1, ys1, zs1, cs1, xs2, ys2, zs2, cs2, data, dnum, free))
			status = ERROR;
	    }
	 if(status != ERROR)
	     {
	      /* add point 0 */
	      i3 = i0;
	      if(x>xs)
                {
		 x = xs1;
		 xs1 = xs2;
                 xs2 = x;
		 x = ys1;
                 ys1 = ys2;
		 ys2 = x;
		 x = zs1;
		 zs1 = zs2;
		 zs2 = x;
                 x = cs1;
		 cs1 = cs2;
                 cs2 = x;
                 l2 = i2;
                 i2 = i1;
                 i1 = l2;
		}
	      else
		l2 = i1;
	      minM += dimy;
	      for(i=0; i<count; i++)
		 {
		  i4 = inside[i];
		  y = *(ma+i4);
		  if(y < minM)
		    {
		     xs = *(xd+i4);
		     ys = *(yd+i4);
		     zs = *(zd+i4);
		     cs = *(col+i4);
		     if(l2 != i4)
                       {
			*(xd+i3) = xs1;
			*(yd+i3) = ys1;
			*(zd+i3) = zs1;
			*(col+i3) = cs1;
			*(p1+i3) = l2;
			*(p2+i3) = i4;
			x = *(mi+i4);
			l1 = *(p1+i4);
			l2 = *(p2+i4);
			delx3 = *(mi+l1)-x;
			dely3 = *(ma+l1)-y;
			delx4 = *(mi+l2)-x;
			dely4 = *(ma+l2)-y;
			if(dely3*delx4<dely4*delx3)
			  {
			   *(p1+i4) = l2;
			   l2 = l1;
			  }
			*(p2+i4) = i3;
                	num++;
                	i3 = num;
			if(*(p1+l2)==i4)
			  *(p1+l2) = i3;
			else
			  *(p2+l2) = i3;
		       }
		     else
		       {
			x = *(mi+i4);
			l1 = *(p1+i4);
			l2 = *(p2+i4);
			if(l2==i3)
			  l2 = l1;
			i3 = i4;
		        num -= 1;
		       }
		     xs1 = xs;
		     ys1 = ys;
		     zs1 = zs;
		     cs1 = cs;
                    }
                 }
              /* insert last */
	      *(xd+i3) = xs1;
	      *(yd+i3) = ys1;
	      *(zd+i3) = zs1;
	      *(col+i3) = cs1;
	      *(p1+i3) = l2;
	      i4 = i3;
	      num++;
	      i3 = num;
	      *(xd+i3) = xs2;
	      *(yd+i3) = ys2;
	      *(zd+i3) = zs2;
	      *(col+i3) = cs2;
	      *(p1+i3) = i4;
	      *(p2+i3) = i2;
	      *(p2+i4) = i3;
	      if(*(p1+i2) == i0)
		*(p1+i2) = i3;
	      else
		*(p2+i2) = i3;
	      num++;
	      *nump = num;
	      /* now sort the data again */
	      for(i=0; i< num; i++)
		 rindex[i] = i;
	      qsort(rindex, num, sizeof(int), O3_sortIndex);
	      /* Create inverse index */
	      for(i=0; i<num;i++)
		 index[rindex[i]]=i;
	     }
         }
      }
    else
      {
       /* no points inside */
       if(O3_InsertTriangle(*(xd+i0), *(yd+i0), *(zd+i0), *(col+i0),
	*(xd+i1), *(yd+i1), *(zd+i1), *(col+i1), *(xd+i2), *(yd+i2), *(zd+i2),
	*(col+i2), data, dnum, free))	status = ERROR;
       /* remove point */
       O3_CLOSE_LINE(p1,p2,i0,i1,i2);
       O3_CLOSE_LINE(p1,p2,i0,i2,i1);
       if(num <= 2)
          *nump = 0;
       else
          *nump = num;
       
      }
    return(status);
   }

/*
 * transform a polygon into a set of triangles
 * data: triangle data (out)
 * free: free values in data
 * xd,yd,zd: input polygon x0,y0,z0 -> xN,yN,zN
 * col:	color indices c0 -> cN
 * num: number of points in poly
 * p1,p2: links to next points in polygon
 * returns: error status
 * 
 */

int O3_Poly2Triangle(double *xd, double *yd, double *zd, double *col,
	int num, int* p1, int *p2, double **data, int *free, int *dnum)
   {
    int i, index[2*MAX_POLY], rindex[2*MAX_POLY], status;
    double maxX, minX, maxY, minY, maxZ, minZ, val;
    double del1, del2, del3;

    /*
     * Maxima and Minima
     * for sorting
     */
    maxX = *xd;
    minX = maxX;
    maxY = *yd;
    minY = maxY;
    maxZ = *zd;
    minZ = maxZ;
    status = OK;
    for(i=1; i<num; i++)
       {
        val = *(xd+i);
        if(val > maxX)
          maxX = val;
        else if(val < minX)
          minX = val;
        val = *(yd+i);
        if(val > maxY)
          maxY = val;
        else if(val < minY)
          minY = val;
        val = *(zd+i);
        if(val > maxZ)
          maxZ = val;
        else if(val < minZ)
          minZ = val;
       }
    /* find out which coordinate is best to sort */
    del1 = maxX-minX;
    del2 = maxY-minY;
    del3 = maxZ-minZ;
    if(del1 > del2)
      {
       if(del1 > del3)
         {
	  Major3D = xd;
          MajDim = del1;
	  if(del2 > del3)
            {
	     Minor3D = yd;
             MinDim = del2;
            }
	  else
            {
	     Minor3D = zd;
             MinDim = del3;
            }
         }
       else
         {
	  Major3D = zd;
          MajDim = del3;
	  if(del1 > del2)
            {
	     Minor3D = xd;
             MinDim = del1;
            }
	  else
            {
	     Minor3D = yd;
             MinDim = del2;
            }
         }
      }
    else
      {
       if(del2 > del3)
         {
	  Major3D = yd;
          MajDim = del2;
	  if(del1 > del3)
            {
	     Minor3D = xd;
             MinDim = del1;
            }
	  else
            {
	     Minor3D = zd;
             MinDim = del3;
            }
         }
       else
         {
	  Major3D = zd;
          MajDim = del3;
	  if(del1 > del2)
            {
	     Minor3D = xd;
             MinDim = del1;
            }
	  else
            {
	     Minor3D = yd;
             MinDim = del2;
            }
         }
      }
    MinDim *= MIN_FEATURE;
    MajDim *= MIN_FEATURE;
    PL1 = p1;
    PL2 = p2;
    /* Now sort the index */
    for(i=0; i< num; i++)
       rindex[i] = i;
    qsort(rindex, num, sizeof(int), O3_sortIndex);
    /* Create inverse index */
    for(i=0; i<num;i++)
       index[rindex[i]]=i;
    while(num)
       {
	if(O3_NextTriangle(Major3D, Minor3D, xd, yd, zd, col, MinDim, MajDim,
		p1, p2, index, rindex, &num, data, dnum, free))
	  {
	   status = ERROR;
           break;
          }
      }
    return(status);
   }

/*
 * simple major minor sort
 */

void O3_SortMM(double *Min, double *Maj, int num, int *points)
   {
    Major3D = Maj;
    Minor3D = Min;
    qsort(points, num, sizeof(int), O3_sortMinMaj);
   }

/*
 * sort bye delx,dely
 */

void O3_SortDxDy(double x0, double y0, double *Min, double *Maj,
	int num, int *points)
   {
    Min0 = x0;
    Maj0 = y0;
    Major3D = Maj;
    Minor3D = Min;
    qsort(points, num, sizeof(int), O3_sortDelxDely);
   }
