/****************************************************************/
/*								*/
/* vrml.c:	Handle VRML files				*/
/*		(at the moment export only)			*/
/*								*/
/*		Design: Walter Benzing 1995			*/
/*								*/
/*		NO RIGHTS RESERVED				*/
/****************************************************************/

/* standard includes */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

/* my includes */

#include "xdim.h"

/*
 * Init VRML output and write header
 */

FILE *VR_InitVRMLout(char *name)
  {
   FILE *out;

   out = fopen(name, "w");
   if(out)
      fprintf(out,"#VRML V1.0 ascii\nDEF XDimData Separator {\n");
   return(out);
  }

/*
 * Write Palettte to VRML
 */

void VR_WritePalette(FILE *out, double ambient, int numCol,
   unsigned char *red, unsigned char *green, unsigned char *blue)
  {
   double fr, fg, fb, fact;
   int i;

   numCol--;
   /* write palette header */
   fprintf(out, "\tMaterial {\n");
   /*
    * write ambient
    */
   fact = ambient/255.0;
   fprintf(out, "\t\tambientColor [\n");
   for(i=0; i<=numCol; i++)
      {
       fr = fact*(double)(red[i]);
       fg = fact*(double)(green[i]);
       fb = fact*(double)(blue[i]);
       if(i!=numCol)
	 fprintf(out, "\t\t\t%f %f %f,\n", fr, fg, fb);
       else
         fprintf(out, "\t\t\t%f %f %f ]\n", fr, fg, fb);
      }
   /*
    * write diffuse
    */
   fact = 1.0/255.0;
   fprintf(out, "\t\tdiffuseColor [\n");
   for(i=0; i<=numCol; i++)
      {
       fr = fact*(double)(red[i]);
       fg = fact*(double)(green[i]);
       fb = fact*(double)(blue[i]);
       if(i!=numCol)
	 fprintf(out, "\t\t\t%f %f %f,\n", fr, fg, fb);
       else
         fprintf(out, "\t\t\t%f %f %f ]\n", fr, fg, fb);
      }
   /* close material node */
   fprintf(out, "\t}\n");
  }

/*
 * write out double z-data on a grid
 */

void VR_WriteDouble(FILE *out, int width, int height, double offx, double factx,
   double offy, double facty, double max, double min,
   double *data, myPixel *deco)
  {
   int i, j, l0, l1, l2, l3, l4, l5, mx, my, last;
   myPixel col;
   double *dp, dz, x0, y0, z0, fact;

   /* write material binding */
   fprintf(out,"\tMaterialBinding { value PER_VERTEX_INDEXED }\n");
   /*
    * write coordinates
    */
   fprintf(out, "\tCoordinate3 {\n\t\tpoint [\n");
   dp = data;
   y0 = offy;
   my = height-1;
   mx = width-1;
   if(min > 0.0 || max < 0.0)
     dz = min;
   else
     dz = 0.0;
   for(i=0; i<height; i++)
      {
       x0 = offx;
       for(j=0; j<width; j++)
          {
           z0 = *dp++;
	   z0 = Mminmax(z0,min,max);
	   z0 -= dz;
           if(i==my && j==mx)
             fprintf(out,"\t\t\t%e %e %e ]\n", x0, y0, z0);
           else
             fprintf(out,"\t\t\t%e %e %e,\n", x0, y0, z0);
           x0 += factx;
          }
       y0 += facty;
      }
   /* end coordinates */
   fprintf(out,"\t}\n"); 
   /*
    * write face set
    */
   fprintf(out,"\tIndexedFaceSet {\n");
   /*
    * coordinate index
    */
   last = width*height-1;
   fprintf(out,"\t\tcoordIndex [\n");
   l1= 0;
   for(i=0; i<my; i++)
      {
       l0 = l1;
       l1 += width;
       l2 = l0;
       l3 = l1;
       for(j=0; j<mx; j++)
          {
           l4 = l2+1;
           l5 = l3+1;
           if(l5 !=last)
             fprintf(out,"\t\t\t%d, %d, %d, -1, %d, %d, %d, -1,\n",
		l2, l3, l5,l2, l4, l5);
	   else
             fprintf(out,"\t\t\t%d, %d, %d, -1, %d, %d, %d, -1 ]\n",
		l2, l3, l5,l2, l4, l5);
	   l2 = l4;
	   l3 = l5; 
          }
      }
   /*
    * Material index
    */
   fprintf(out,"\t\tmaterialIndex [\n");
   if(max > min)
     fact = (double)(CO_ONE-1)/(max - min);
   else
     fact = 1.0;
   dp = data;
   for(i=0; i<= last; i++)
      {
       z0 = *dp++;
       z0 = Mminmax(z0,min,max);
       l0 = (int)(fact*(z0-min));
       col = deco[l0];
       if(i!=last)
         fprintf(out, "\t\t\t%u,\n", col);
       else
         fprintf(out, "\t\t\t%u ]\n", col);
      }
   /* end face set */
   fprintf(out,"\t}\n");
  }

/*
 * Close VRML output
 */

void VR_CloseVRMLout(FILE *out)
  {
   fprintf(out,"}\n");
   fclose(out);
  }
