/*
 * xdjpeg.c
 * 
 * exporting pictures from xdim as jpeg's with the JPEG library
 * from the Independent JPEG Group.
 *          ======================
 *
 * example.c from the jepegsrc.v5a was used as the skeleton
 * read the README in the jpeg directory file about use and copyright
 *          ======
 */


#include <stdlib.h>
#include <setjmp.h>

/* my includes (W.B.) */

#include "xdim.h"

/*
 * Include file for users of JPEG library.
 */

#include "jpeglib.h"
#include "jerror.h"

/* Create the add-on message string table. */
/* Boy this really gave me a hard time */
/* Who ever thought this thing out, should have also thought of a documentation*/

#include "xdjperr.h"

struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */

  jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;


/******************** JPEG COMPRESSION SAMPLE INTERFACE *******************/

/* This half of the example shows how to feed data into the JPEG compressor.
 * We present a minimal version that does not worry about refinements such
 * as error recovery (the JPEG code will just exit() if it gets an error).
 */


/*
 * IMAGE DATA FORMATS:
 *
 * The standard input image format is a rectangular array of pixels, with
 * each pixel having the same number of "component" values (color channels).
 * Each pixel row is an array of JSAMPLEs (which typically are unsigned chars).
 * If you are working with color data, then the color values for each pixel
 * must be adjacent in the row; for example, R,G,B,R,G,B,R,G,B,... for 24-bit
 * RGB color.
 *
 */

/*
 * My Routine for message output
 */

METHODDEF void
JP_output_message (j_common_ptr cinfo)
{
  extern XDimInfo XDim;

  /* Create the message */
  (*cinfo->err->format_message) (cinfo, XDim.error);

  /* send it to the error window*/
  WD_XDimError(&XDim, NULL);
}


/*
 * Here's the routine that will replace the standard error_exit method:
 */

METHODDEF void
JP_error_exit (j_common_ptr cinfo)
{
  /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
  my_error_ptr myerr = (my_error_ptr) cinfo->err;

  /* Always display the message. */
  /* We could postpone this until after returning, if we chose. */
  (*cinfo->err->output_message) (cinfo);

  /* Return control to the setjmp point */
  longjmp(myerr->setjmp_buffer, 1);
}

/*
 * translate internal format at line to jpeg format
 */

void JP_TranslateNext(JSAMPLE *image, int line, ViewportInfo *view
	, XDimInfo *XDim)
   {
    XDHardwareInfo *HWI;
    int i, num;
    unsigned char *red, *green, *blue;
    myPixel *mp2jp, pix;
    
    HWI = &(XDim->hwInfo);
    num = view->pwidth;
    red = HWI->red;
    green = HWI->green;
    blue = HWI->blue;
    mp2jp = view->pict+line*num;
    for(i=0; i<num; i++)
       {
	pix = *mp2jp++;
	*image++ = (JSAMPLE) red[pix];		/* red */
	*image++ = (JSAMPLE) green[pix];	/* green */
	*image++ = (JSAMPLE) blue[pix];		/* blue */
       }
   }

/*
 * Sample routine for JPEG compression.  We assume that the target file name
 * and a compression quality factor are passed in.
 */

int JP_write_JPEG_file (char * filename, XDimInfo *XDim)
{
  /* This struct contains the JPEG compression parameters and pointers to
   * working space (which is allocated as needed by the JPEG library).
   * It is possible to have several such structures, representing multiple
   * compression/decompression processes, in existence at once.  We refer
   * to any one struct (and its associated working data) as a "JPEG object".
   */
  struct jpeg_compress_struct cinfo;
  /* This struct represents a JPEG error handler.  It is declared separately
   * because applications often want to supply a specialized error handler
   * (see the second half of this file for an example).  But here we just
   * take the easy way out and use the standard error handler, which will
   * print a message on stderr and call exit() if compression fails.
   */
  struct my_error_mgr jerr;
  /* More stuff */
  FILE *outfile=NULL;		/* target file */
  JSAMPLE *image_buffer;	/* Points to large array of R,G,B-order data */
  JSAMPROW row_pointer[1];	/* pointer to JSAMPLE row[s] */

  /*
   * Allocate a Viewport (W.B.)
   */
  ViewportInfo *view;

  if(XDim->port <0)
    {
     strcpy(XDim->error, XDERR_NO_VIEW);
     return(ERROR);
    }
   view = &((XDim->Views)[XDim->port]);
   if(view->pict == NULL)
     {
      strcpy(XDim->error, XDERR_NO_IMAGE);
      return(ERROR);
     }
   /*
    * Allocate a buffer
    */
   image_buffer = malloc(3*view->pwidth);
   if(!image_buffer)
     {
      sprintf(XDim->error, XDERR_NO_MEM_D, 3*view->pwidth);
      return(ERROR);
     }
  /* Step 1: allocate and initialize JPEG compression object */

  /* We have to set up the error handler first, in case the initialization
   * step fails.  (Unlikely, but it could happen if you are out of memory.)
   * This routine fills in the contents of struct jerr, and returns jerr's
   * address which we place into the link field in cinfo.
   */
  cinfo.err = jpeg_std_error(&jerr.pub);
  jerr.pub.error_exit = JP_error_exit;
  jerr.pub.output_message = JP_output_message;
  jerr.pub.addon_message_table = xdjpeg_message_table;
  jerr.pub.first_addon_message = JMSG_FIRSTADDONCODE;
  jerr.pub.last_addon_message = JMSG_LASTADDONCODE;
  /* Establish the setjmp return context for my_error_exit to use. */
  if (setjmp(jerr.setjmp_buffer)) {
    /* If we get here, the JPEG code has signaled an error.
     * We need to clean up the JPEG object, close the input file, and return.
     */
    jpeg_destroy_compress(&cinfo);
    fclose(outfile);
    free(image_buffer);
    return(ERROR);
  }
  /* Now we can initialize the JPEG compression object. */
  jpeg_create_compress(&cinfo);

  /* Step 2: specify data destination (eg, a file) */
  /* Note: steps 2 and 3 can be done in either order. */

  /* Here we use the library-supplied code to send compressed data to a
   * stdio stream.  You can also write your own code to do something else.
   * VERY IMPORTANT: use "b" option to fopen() if you are on a machine that
   * requires it in order to write binary files.
   */
  if ((outfile = fopen(filename, "wb")) == NULL) {
     sprintf(XDim->error, XDERR_OPEN_WRITE_S, filename);
     return(ERROR);
  }
  jpeg_stdio_dest(&cinfo, outfile);

  /* Step 3: set parameters for compression */

  /* First we supply a description of the input image.
   * Four fields of the cinfo struct must be filled in:
   */
  cinfo.image_width = view->pwidth; 	/* image width and height, in pixels */
  cinfo.image_height = view->pheight;
  cinfo.input_components = 3;		/* # of color components per pixel */
  cinfo.in_color_space = JCS_RGB; 	/* colorspace of input image */
  /* Now use the library's routine to set default compression parameters.
   * (You must set at least cinfo.in_color_space before calling this,
   * since the defaults depend on the source color space.)
   */
  jpeg_set_defaults(&cinfo);
  /* Now you can set any non-default parameters you wish to.
   * Here we just illustrate the use of quality (quantization table) scaling:
   */
  jpeg_set_quality(&cinfo, (XDim->colors).jpgQuality, TRUE /* limit to baseline-JPEG values */);

  /* Step 4: Start compressor */

  /* TRUE ensures that we will write a complete interchange-JPEG file.
   * Pass TRUE unless you are very sure of what you're doing.
   */
  jpeg_start_compress(&cinfo, TRUE);

  /* Step 5: while (scan lines remain to be written) */
  /*           jpeg_write_scanlines(...); */

  /* Here we use the library's state variable cinfo.next_scanline as the
   * loop counter, so that we don't have to keep track ourselves.
   * To keep things simple, we pass one scanline per call; you can pass
   * more if you wish, though.
   */

  while (cinfo.next_scanline < cinfo.image_height) {
   JP_TranslateNext(image_buffer, cinfo.next_scanline, view, XDim);
    row_pointer[0] = image_buffer;
    (void) jpeg_write_scanlines(&cinfo, row_pointer, 1);
  }

  /* Step 6: Finish compression */

  jpeg_finish_compress(&cinfo);
  /* After finish_compress, we can close the output file. */
  fclose(outfile);
  /* release image buffer */
  free(image_buffer);

  /* Step 7: release JPEG compression object */

  /* This is an important step since it will release a good deal of memory. */
  jpeg_destroy_compress(&cinfo);

  /* And we're done! */
 return(OK);
}


/*
 * SOME FINE POINTS:
 *
 * In the above loop, we ignored the return value of jpeg_write_scanlines,
 * which is the number of scanlines actually written.  We could get away
 * with this because we were only relying on the value of cinfo.next_scanline,
 * which will be incremented correctly.  If you maintain additional loop
 * variables then you should be careful to increment them properly.
 * Actually, for output to a stdio stream you needn't worry, because
 * then jpeg_write_scanlines will write all the lines passed (or else exit
 * with a fatal error).  Partial writes can only occur if you use a data
 * destination module that can demand suspension of the compressor.
 * (If you don't know what that's for, you don't need it.)
 *
 * If the compressor requires full-image buffers (for entropy-coding
 * optimization or a noninterleaved JPEG file), it will create temporary
 * files for anything that doesn't fit within the maximum-memory setting.
 * (Note that temp files are NOT needed if you use the default parameters.)
 * On some systems you may need to set up a signal handler to ensure that
 * temporary files are deleted if the program is interrupted.  See libjpeg.doc.
 *
 * Scanlines MUST be supplied in top-to-bottom order if you want your JPEG
 * files to be compatible with everyone else's.  If you cannot readily read
 * your data in that order, you'll need an intermediate array to hold the
 * image.  See rdtarga.c or rdbmp.c for examples of handling bottom-to-top
 * source data using the JPEG code's internal virtual-array mechanisms.
 */



/******************** JPEG DECOMPRESSION SAMPLE INTERFACE *******************/

/* This half of the example shows how to read data from the JPEG decompressor.
 * It's a bit more refined than the above, in that we show:
 *   (a) how to modify the JPEG library's standard error-reporting behavior;
 *   (b) how to allocate workspace using the library's memory manager.
 *
 * Just to make this example a little different from the first one, we'll
 * assume that we do not intend to put the whole image into an in-memory
 * buffer, but to send it line-by-line someplace else.  We need a one-
 * scanline-high JSAMPLE array as a work buffer, and we will let the JPEG
 * memory manager allocate it for us.  This approach is actually quite useful
 * because we don't need to remember to deallocate the buffer separately: it
 * will go away automatically when the JPEG object is cleaned up.
 */


/*
 * ERROR HANDLING:
 *
 * The JPEG library's standard error handler (jerror.c) is divided into
 * several "methods" which you can override individually.  This lets you
 * adjust the behavior without duplicating a lot of code, which you might
 * have to update with each future release.
 *
 * Our example here shows how to override the "error_exit" method so that
 * control is returned to the library's caller when a fatal error occurs,
 * rather than calling exit() as the standard error_exit method does.
 *
 * We use C's setjmp/longjmp facility to return control.  This means that the
 * routine which calls the JPEG library must first execute a setjmp() call to
 * establish the return point.  We want the replacement error_exit to do a
 * longjmp().  But we need to make the setjmp buffer accessible to the
 * error_exit routine.  To do this, we make a private extension of the
 * standard JPEG error handler object.  (If we were using C++, we'd say we
 * were making a subclass of the regular error handler.)
 */

static JSAMPLE *IMred, *IMgreen, *IMblue;
static int *JPindex;

/*
 * Sort colors to reduce jitter (W.B.)
 */

int JP_sortColor(const void *index1,const void *index2)
  {
   int r1, g1, b1, r2, g2, b2;
   int i1, i2;
   myCoor ret, ret1, ret2;
   const int *ind1, *ind2;
   
   ind1 = index1;
   ind2 = index2;
   i1 = *ind1;
   i2 = *ind2;
   r1 = GETJSAMPLE(IMred[i1]);
   g1 = GETJSAMPLE(IMgreen[i1]);
   b1 = GETJSAMPLE(IMblue[i1]);
   r2 = GETJSAMPLE(IMred[i2]);
   g2 = GETJSAMPLE(IMgreen[i2]);
   b2 = GETJSAMPLE(IMblue[i2]);
   ret1 = GR_MapIndex(r1, g1, b1);
   ret2 = GR_MapIndex(r2, g2, b2);
   ret = ret1 -ret2;
   return(ret);
  }

/*
 * Process the color map to reduce noise
 */

void
JP_put_color_map (struct jpeg_decompress_struct *cinfo, XDimInfo *XDim)
/* Write the color map */
{
  XDHardwareInfo *HWI;
  XDColorInfo *XDC;
  int i, rows, shiftL, shiftR, *ind, ind1, ind2, *NextInd, num_colors;
  unsigned short dum;

  HWI = &(XDim->hwInfo);
  XDC = &(XDim->colors);
  num_colors = cinfo->actual_number_of_colors;
  if(num_colors > 0)
    {
     shiftL = (sizeof(short)<<CHAR2BITS)- cinfo->data_precision; 
     shiftR = cinfo->data_precision - shiftL;
     GR_InitColorImport( num_colors, num_colors, XDC, HWI);
     if((JPindex= malloc(sizeof(int)*num_colors)) == NULL)
       {
	ERREXIT1(cinfo, JERR_IMP_MEMLOST, num_colors);
       }
     rows = 0;
     ind = JPindex;
     for(i= 0; i<num_colors; i++)
        *ind++ = i;
     if(cinfo->out_color_space != JCS_GRAYSCALE)
       {
        IMred = cinfo->colormap[0];
        IMgreen = cinfo->colormap[1];
        IMblue = cinfo->colormap[2];
        qsort(JPindex, num_colors, sizeof(int), JP_sortColor);
	for(i=0; i<num_colors >>1; i++)
           {
            (XDC->colors)[i] = 2;
            ind1 = JPindex[i<<1];
            ind2 = JPindex[(i<<1)+1];
	    dum =  GETJSAMPLE(IMred[ind1]);
       	    (XDC->beginR)[i] = (dum<<shiftL)|(dum>>shiftR);
	    dum =  GETJSAMPLE(IMgreen[ind1]);
       	    (XDC->beginG)[i] = (dum<<shiftL)|(dum>>shiftR);
	    dum =  GETJSAMPLE(IMblue[ind1]);
       	    (XDC->beginB)[i] = (dum<<shiftL)|(dum>>shiftR);
	    dum =  GETJSAMPLE(IMred[ind2]);
       	    (XDC->endR)[i] = (dum<<shiftL)|(dum>>shiftR);
	    dum =  GETJSAMPLE(IMgreen[ind2]);
       	    (XDC->endG)[i] = (dum<<shiftL)|(dum>>shiftR);
	    dum =  GETJSAMPLE(IMblue[ind2]);
       	    (XDC->endB)[i] = (dum<<shiftL)|(dum>>shiftR);
       	    rows++;
           }
	if(num_colors &1)
	  {
           (XDC->colors)[i] = 2;
           ind2 = JPindex[i<<1];
	   dum =  GETJSAMPLE(IMred[ind2]);
       	   (XDC->beginR)[i] = (dum<<shiftL)|(dum>>shiftR);
	   dum =  GETJSAMPLE(IMgreen[ind2]);
       	   (XDC->beginG)[i] = (dum<<shiftL)|(dum>>shiftR);
	   dum =  GETJSAMPLE(IMblue[ind2]);
       	   (XDC->beginB)[i] = (dum<<shiftL)|(dum>>shiftR);
	   dum =  GETJSAMPLE(IMred[ind2]);
       	   (XDC->endR)[i] = (dum<<shiftL)|(dum>>shiftR);
	   dum =  GETJSAMPLE(IMgreen[ind2]);
       	   (XDC->endG)[i] = (dum<<shiftL)|(dum>>shiftR);
	   dum =  GETJSAMPLE(IMblue[ind2]);
       	   (XDC->endB)[i] = (dum<<shiftL)|(dum>>shiftR);
	   rows++;
	  }
        if((NextInd= malloc(sizeof(int)*num_colors)) == NULL)
	  {
	   ERREXIT1(cinfo, JERR_IMP_MEMLOST, num_colors);
	  }
        for(i=0; i<num_colors; i++)
           NextInd[JPindex[i]] = i;
        free(JPindex);
        JPindex = NextInd;              
       }
     else     
       {
        i = 0;
        (XDC->colors)[i] = num_colors;
        if(GETJSAMPLE(cinfo->colormap[0][0])< GETJSAMPLE(cinfo->colormap[0][1]))
          {
       	   (XDC->beginR)[i] = 0;
       	   (XDC->beginG)[i] = 0;
       	   (XDC->beginB)[i] = 0;
       	   (XDC->endR)[i] = 65535;
       	   (XDC->endG)[i] = 65535;
       	   (XDC->endB)[i] = 65535;
       	  }
       	else
          {
       	   (XDC->beginR)[i] = 65535;
       	   (XDC->beginG)[i] = 65535;
       	   (XDC->beginB)[i] = 65535;
       	   (XDC->endR)[i] = 0;
       	   (XDC->endG)[i] = 0;
       	   (XDC->endB)[i] = 0;
       	  }
       	rows++;
       }
     XDC->rowNum = rows;
    }
}

/*
 * Sample routine for JPEG decompression.  We assume that the source file name
 * is passed in.  We want to return 1 on success, 0 on error.
 */


int JP_read_JPEG_file (XDimInfo *XDim)
{
  /* This struct contains the JPEG decompression parameters and pointers to
   * working space (which is allocated as needed by the JPEG library).
   */
  struct jpeg_decompress_struct cinfo;
  /* We use our private extension JPEG error handler. */
  struct my_error_mgr jerr;
  /* More stuff */
  FILE * infile;		/* source file */
  JSAMPARRAY buffer;		/* Output row buffer */
  JSAMPROW ptr0;
  int row_stride;		/* physical row width in output buffer */
  int col;			/* counter */
  unsigned char *jp2cp;		/* data pointer */
  ViewportInfo *view;

  /* In this example we want to open the input file before doing anything else,
   * so that the setjmp() error recovery below can assume the file is open.
   * VERY IMPORTANT: use "b" option to fopen() if you are on a machine that
   * requires it in order to read binary files.
   */

 if(XDim->port <0)
   {
    strcpy(XDim->error, XDERR_NO_VIEW);
    return(ERROR);
   }
  view = &((XDim->Views)[XDim->port]);
  if ((infile = fopen(view->fileName, "rb")) == NULL) {
     sprintf(XDim->error, XDERR_OPEN_READ_S, view->fileName);
     return(ERROR);
  }

  /* Step 1: allocate and initialize JPEG decompression object */

  /* We set up the normal JPEG error routines, then override error_exit. */
  cinfo.err = jpeg_std_error(&jerr.pub);
  jerr.pub.error_exit = JP_error_exit;
  jerr.pub.output_message = JP_output_message;
  jerr.pub.addon_message_table = xdjpeg_message_table;
  jerr.pub.first_addon_message = JMSG_FIRSTADDONCODE;
  jerr.pub.last_addon_message = JMSG_LASTADDONCODE;
  /* Establish the setjmp return context for my_error_exit to use. */
  if (setjmp(jerr.setjmp_buffer)) {
    /* If we get here, the JPEG code has signaled an error.
     * We need to clean up the JPEG object, close the input file, and return.
     */
    jpeg_destroy_decompress(&cinfo);
    fclose(infile);
    return(ERROR);
  }
  /* Now we can initialize the JPEG decompression object. */
  jpeg_create_decompress(&cinfo);

  /* Step 2: specify data source (eg, a file) */

  jpeg_stdio_src(&cinfo, infile);

  /* Step 3: read file parameters with jpeg_read_header() */

  (void) jpeg_read_header(&cinfo, TRUE);
  /* We can ignore the return value from jpeg_read_header since
   *   (a) suspension is not possible with the stdio data source, and
   *   (b) we passed TRUE to reject a tables-only JPEG file as an error.
   * See libjpeg.doc for more info.
   */

  /* Step 4: set parameters for decompression */

  cinfo.data_precision = 8;		/* bits per pixel component value */
  cinfo.quantize_colors = TRUE;
  cinfo.desired_number_of_colors = MAX_COL_NUMBER;

  /* Step 5: Start decompressor */

  jpeg_start_decompress(&cinfo);

  /* We may need to do some setup of our own at this point before reading
   * the data.  After jpeg_start_decompress() we have the correct scaled
   * output image dimensions available, as well as the output colormap
   * if we asked for color quantization.
   * In this example, we need to make an output work buffer of the right size.
   */ 

  /* Allocate memory in a viewport (W.B.) */

  if(CO_NewData(cinfo.image_width, cinfo.image_height, MY_UCHAR, XDim))
    {
     ERREXIT2(&cinfo, JERR_IMP_PIXMEM, cinfo.image_width, cinfo.image_height);
    }
  view->Width = cinfo.image_width;		/* width */
  view->Height = cinfo.image_height;		/* height */
   /*
    * Process the palette
    */
  JP_put_color_map(&cinfo, XDim);

  jp2cp = view->charData;			/* data */
  /* JSAMPLEs per row in output buffer */
  row_stride = cinfo.output_width * cinfo.output_components;
  /* Make a one-row-high sample array that will go away when done with image */
  buffer = (*cinfo.mem->alloc_sarray)
		((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

  /* Step 6: while (scan lines remain to be read) */
  /*           jpeg_read_scanlines(...); */

  /* Here we use the library's state variable cinfo.output_scanline as the
   * loop counter, so that we don't have to keep track ourselves.
   */
  while (cinfo.output_scanline < cinfo.output_height) {
    (void) jpeg_read_scanlines(&cinfo, buffer, 1);
    /* Assume put_scanline_someplace wants a pointer and sample count. */
    ptr0 = buffer[0];
    for (col = 0; col < row_stride; col++) {
        *jp2cp++ = (unsigned char)JPindex[(int)(GETJSAMPLE(*ptr0))];
        ptr0++;
       }
  }

  /* Step 7: Finish decompression */

  (void) jpeg_finish_decompress(&cinfo);
  /* We can ignore the return value since suspension is not possible
   * with the stdio data source.
   */

  /* Step 8: Release JPEG decompression object */

  /* This is an important step since it will release a good deal of memory. */
  jpeg_destroy_decompress(&cinfo);

  /* After finish_decompress, we can close the input file.
   * Here we postpone it until after no more JPEG errors are possible,
   * so as to simplify the setjmp error logic above.  (Actually, I don't
   * think that jpeg_destroy can do an error exit, but why assume anything...)
   */
  fclose(infile);

  /* At this point you may want to check to see whether any corrupt-data
   * warnings occurred (test whether jerr.pub.num_warnings is nonzero).
   */
  /*
   * set the parameters (W.B.)
   */

  strcpy(view->DimX, "pixel");
  strcpy(view->DimY, "pixel");
  strcpy(view->DimZ, "color");
  view->FactX = 1.0;
  view->FactY = 1.0;
  view->OffX = 0.0;
  view->OffY = 0.0;
  view->links = 0;

  return(OK);			/* indicate success */

  /* And we're done! */
  return 1;
}


/*
 * SOME FINE POINTS:
 *
 * In the above code, we ignored the return value of jpeg_read_scanlines,
 * which is the number of scanlines actually read.  We could get away with
 * this because we asked for only one line at a time and we weren't using
 * a suspending data source.  See libjpeg.doc for more info.
 *
 * We cheated a bit by calling alloc_sarray() after jpeg_start_decompress();
 * we should have done it beforehand to ensure that the space would be
 * counted against the JPEG max_memory setting.  In some systems the above
 * code would risk an out-of-memory error.  However, in general we don't
 * know the output image dimensions before jpeg_start_decompress(), unless we
 * call jpeg_calc_output_dimensions().  See libjpeg.doc for more about this.
 *
 * Scanlines are returned in the same order as they appear in the JPEG file,
 * which is standardly top-to-bottom.  If you must emit data bottom-to-top,
 * you can use one of the virtual arrays provided by the JPEG memory manager
 * to invert the data.  See wrbmp.c for an example.
 *
 * As with compression, some operating modes may require temporary files.
 * On some systems you may need to set up a signal handler to ensure that
 * temporary files are deleted if the program is interrupted.  See libjpeg.doc.
 */
